from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

from .literals import (
    EVENT_EVENTS_CLEARED_NAME, EVENT_EVENTS_EXPORTED_NAME,
    EVENT_TYPE_NAMESPACE_NAME
)

namespace = EventTypeNamespace(
    label=_('Events'), name=EVENT_TYPE_NAMESPACE_NAME
)

event_events_cleared = namespace.add_event_type(
    label=_('Events cleared'), name=EVENT_EVENTS_CLEARED_NAME
)
event_events_exported = namespace.add_event_type(
    label=_('Events exported'), name=EVENT_EVENTS_EXPORTED_NAME
)



from django.conf.urls import url

from .views import (
    UserOTPDataDetailView, UserOTPDataDisableView, UserOTPDataEnableView,
    UserOTPDataVerifyTokenView
)

urlpatterns = [
    url(
        regex=r'^otp/$', name='otp_detail',
        view=UserOTPDataDetailView.as_view()
    ),
    url(
        regex=r'^otp/disable/$', name='otp_disable',
        view=UserOTPDataDisableView.as_view()
    ),
    url(
        regex=r'^otp/enable/$', name='otp_enable',
        view=UserOTPDataEnableView.as_view()
    ),
    url(
        regex=r'^otp/verify/$', name='otp_verify',
        view=UserOTPDataVerifyTokenView.as_view()
    )
]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('converter', '0009_auto_20150714_2228'),
    ]

    operations = [
        migrations.AlterField(
            model_name='transformation',
            name='name',
            field=models.CharField(
                max_length=128, verbose_name='Name',
                choices=[
                    ('rotate', 'Rotate: degrees'), ('zoom', 'Zoom: percent'),
                    ('resize', 'Resize: width, height'),
                    ('crop', 'Crop: left, top, right, bottom')
                ]
            ),
            preserve_default=True,
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(
    label=_('Download files'), name='download_files'
)

event_download_file_created = namespace.add_event_type(
    label=_('Download file created'), name='created'
)
event_download_file_deleted = namespace.add_event_type(
    label=_('Download file deleted'), name='deleted'
)
event_download_file_downloaded = namespace.add_event_type(
    label=_('Download file downloaded'), name='downloaded'
)



from django.utils.translation import ugettext_lazy as _

from drf_yasg import openapi

import mayan

from .literals import API_VERSION

openapi_info_default_version = 'v{}'.format(API_VERSION)
openapi_info_title = _('%s API') % mayan.__title__
openapi_license = openapi.License(name=mayan.__license__)

openapi_info = openapi.Info(
    default_version=openapi_info_default_version,
    description=mayan.__description__, license=openapi_license,
    title=openapi_info_title
)



import pytz

from django.utils import timezone, translation
from django.utils.deprecation import MiddlewareMixin


class UserLocaleProfileMiddleware(MiddlewareMixin):
    def process_request(self, request):
        user = getattr(request, 'user', None)
        if user:
            locale_profile = getattr(user, 'locale_profile', None)
            if locale_profile:
                if locale_profile.language:
                    translation.activate(language=locale_profile.language)

                if locale_profile.timezone:
                    timezone.activate(
                        timezone=pytz.timezone(zone=locale_profile.timezone)
                    )
                else:
                    timezone.deactivate()



from mayan.apps.dynamic_search.tests.mixins.base import SearchTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..permissions import permission_message_view
from ..search import search_model_message

from .mixins import MessageTestMixin


class MessageSearchTestCase(
    MessageTestMixin, SearchTestMixin, BaseTestCase
):
    _test_search_model = search_model_message

    def test_search_model_message_body_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={'body': self._test_message.body}
        )
        self.assertTrue(self._test_message not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_message_body_with_access(self):
        self.grant_access(
            obj=self._test_message, permission=permission_message_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={'body': self._test_message.body}
        )
        self.assertTrue(self._test_message in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_message_date_year_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={
                'date_time': '>{}'.format(
                    self._test_message.date_time.year - 1
                )
            }
        )
        self.assertTrue(self._test_message not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_message_date_year_with_access(self):
        self.grant_access(
            obj=self._test_message, permission=permission_message_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={
                'date_time': '>{}'.format(
                    self._test_message.date_time.year - 1
                )
            }
        )
        self.assertTrue(self._test_message in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_message_subject_no_permission(self):
        self._clear_events()

        queryset = self._do_test_search(
            query={'subject': self._test_message.subject}
        )
        self.assertTrue(self._test_message not in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_search_model_message_subject_with_access(self):
        self.grant_access(
            obj=self._test_message, permission=permission_message_view
        )

        self._clear_events()

        queryset = self._do_test_search(
            query={'subject': self._test_message.subject}
        )
        self.assertTrue(self._test_message in queryset)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.contrib import admin

from mptt.admin import MPTTModelAdmin

from .models import Cabinet


@admin.register(Cabinet)
class CabinetAdmin(MPTTModelAdmin):
    filter_horizontal = ('documents',)
    list_display = ('label',)



from django.db import migrations, models

from ..compatibility import RGBColorField


class Migration(migrations.Migration):
    dependencies = [
        ('tags', '0007_auto_20170118_1758')
    ]

    operations = [
        migrations.AlterField(
            model_name='tag',
            name='color',
            field=RGBColorField(
                help_text='The RGB color values for the tag.',
                verbose_name='Color'
            )
        ),
        migrations.AlterField(
            model_name='tag',
            name='label',
            field=models.CharField(
                db_index=True, help_text='A short text used as the tag '
                'name.', max_length=128, unique=True, verbose_name='Label'
            )
        )
    ]



from django.utils.translation import ugettext_lazy as _

SOURCE_UNCOMPRESS_CHOICE_ALWAYS = 'y'
SOURCE_UNCOMPRESS_CHOICE_NEVER = 'n'
SOURCE_UNCOMPRESS_CHOICE_ASK = 'a'

SOURCE_UNCOMPRESS_NON_INTERACTIVE_CHOICES = (
    (SOURCE_UNCOMPRESS_CHOICE_ALWAYS, _('Always')),
    (SOURCE_UNCOMPRESS_CHOICE_NEVER, _('Never'))
)

SOURCE_UNCOMPRESS_INTERACTIVE_CHOICES = (
    (SOURCE_UNCOMPRESS_CHOICE_ALWAYS, _('Always')),
    (SOURCE_UNCOMPRESS_CHOICE_NEVER, _('Never')),
    (SOURCE_UNCOMPRESS_CHOICE_ASK, _('Ask user'))
)



from django.core.exceptions import ValidationError

from mayan.apps.documents.permissions import permission_document_view
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..events import (
    event_cabinet_created, event_cabinet_deleted,
    event_cabinet_document_added, event_cabinet_document_removed,
    event_cabinet_edited
)
from ..models import Cabinet

from .literals import TEST_CABINET_LABEL, TEST_CABINET_LABEL_EDITED
from .mixins import CabinetTestMixin


class CabinetTestCase(CabinetTestMixin, BaseTestCase):
    def test_cabinet_create(self):
        self._clear_events()

        self._create_test_cabinet()

        self.assertEqual(
            Cabinet.objects.all().count(), 1
        )
        self.assertQuerysetEqual(
            Cabinet.objects.all(), (
                repr(self._test_cabinet),
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_cabinet)
        self.assertEqual(events[0].target, self._test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_created.id)

    def test_cabinet_delete(self):
        self._create_test_cabinet()

        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        self._test_cabinet.delete()

        self.assertEqual(
            Cabinet.objects.count(), test_cabinet_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_duplicate_create(self):
        self._create_test_cabinet()

        self._clear_events()

        with self.assertRaises(expected_exception=ValidationError):
            cabinet_2 = Cabinet(
                label=self._test_cabinet_list[0].label
            )
            cabinet_2.validate_unique()
            cabinet_2.save()

        self.assertEqual(
            Cabinet.objects.all().count(), 1
        )
        self.assertQuerysetEqual(
            Cabinet.objects.all(), (
                repr(self._test_cabinet),
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_cabinet_edit(self):
        self._create_test_cabinet()

        self._clear_events()

        test_cabinet_label = self._test_cabinet.label

        self._test_cabinet.label = TEST_CABINET_LABEL_EDITED
        self._test_cabinet.save()

        self.assertNotEqual(self._test_cabinet.label, test_cabinet_label)

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_cabinet)
        self.assertEqual(events[0].target, self._test_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_edited.id)

    def test_cabinet_child_create(self):
        self._create_test_cabinet()

        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        inner_cabinet = Cabinet.objects.create(
            parent=self._test_cabinet, label=TEST_CABINET_LABEL
        )

        self.assertEqual(
            Cabinet.objects.count(), test_cabinet_count + 1
        )
        self.assertQuerysetEqual(
            Cabinet.objects.all(),
            map(
                repr, (self._test_cabinet, inner_cabinet)
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, inner_cabinet)
        self.assertEqual(events[0].target, inner_cabinet)
        self.assertEqual(events[0].verb, event_cabinet_created.id)

    def test_cabinet_child_delete(self):
        self._create_test_cabinet()
        self._create_test_cabinet_child()

        test_cabinet_count = Cabinet.objects.count()

        self._clear_events()

        self._test_cabinet_child.delete()

        self.assertEqual(
            Cabinet.objects.count(), test_cabinet_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, self._test_cabinet)
        self.assertEqual(events[0].target, None)
        self.assertEqual(events[0].verb, event_cabinet_deleted.id)

    def test_method_get_absolute_url(self):
        self._create_test_cabinet()

        self._clear_events()

        self.assertTrue(
            self._test_cabinet.get_absolute_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class CabinetDocumentTestCase(
    CabinetTestMixin, DocumentTestMixin, BaseTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()

    def test_addition_of_documents(self):
        test_cabinet_document_count = self._test_cabinet.documents.count()

        self._clear_events()

        self._test_cabinet.document_add(
            document=self._test_document, user=self._test_case_user
        )

        self.assertEqual(
            self._test_cabinet.documents.count(),
            test_cabinet_document_count + 1
        )
        self.assertQuerysetEqual(
            self._test_cabinet.documents.all(), (
                repr(self._test_document),
            )
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_added.id)

    def test_addition_and_deletion_of_documents(self):
        self._test_cabinet.document_add(
            document=self._test_document, user=self._test_case_user
        )

        test_cabinet_document_count = self._test_cabinet.documents.count()

        self._clear_events()

        self._test_cabinet.document_remove(
            document=self._test_document, user=self._test_case_user
        )

        self.assertEqual(
            self._test_cabinet.documents.count(),
            test_cabinet_document_count - 1
        )
        self.assertQuerysetEqual(
            self._test_cabinet.documents.all(), ()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_cabinet)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_cabinet_document_removed.id)

    def test_cabinet_get_document_count_method(self):
        self._test_cabinet.document_add(
            document=self._test_document, user=self._test_case_user
        )

        test_cabinet_document_count = self._test_cabinet.documents.count()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        self.assertEqual(
            self._test_cabinet.get_document_count(user=self._test_case_user),
            test_cabinet_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_cabinet_document_count_method(self):
        self._test_cabinet.document_add(
            document=self._test_document, user=self._test_case_user
        )

        test_cabinet_document_count = self._test_cabinet.documents.count()

        self._test_document.delete()

        self.grant_access(
            obj=self._test_document, permission=permission_document_view
        )

        self._clear_events()

        self.assertEqual(
            self._test_cabinet.get_document_count(user=self._test_case_user),
            test_cabinet_document_count - 1,
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.core.management.base import BaseCommand

from ...tasks import task_reindex_backend


class Command(BaseCommand):
    help = 'Erases and populates the search backend internal indexes.'

    def handle(self, *args, **options):
        task_reindex_backend.apply_async()



from mayan.apps.common.tests.literals import TEST_ARCHIVE_MSG_STRANGE_DATE_PATH
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.tests.literals import TEST_FILE_PDF_FILENAME

from .literals import (
    TEST_MSG_FILE_METADATA_DOTTED_NAME_SUBJECT,
    TEST_MSG_FILE_METADATA_DOTTED_NAME_TO,
    TEST_MSG_FILE_METADATA_VALUE_SUBJECT, TEST_MSG_FILE_METADATA_VALUE_TO,
    TEST_PDF_FILE_METADATA_DOTTED_NAME, TEST_PDF_FILE_METADATA_VALUE
)


class EXIFToolDriverTestCase(GenericDocumentTestCase):
    _test_document_filename = TEST_FILE_PDF_FILENAME

    def test_driver_entries(self):
        self._test_document.submit_for_file_metadata_processing()

        value = self._test_document.file_latest.get_file_metadata(
            dotted_name=TEST_PDF_FILE_METADATA_DOTTED_NAME
        )
        self.assertEqual(value, TEST_PDF_FILE_METADATA_VALUE)


class ExtractMSGDriverTestCase(GenericDocumentTestCase):
    _test_document_filename = TEST_ARCHIVE_MSG_STRANGE_DATE_PATH

    def test_driver_entries(self):
        self._test_document.submit_for_file_metadata_processing()

        value_subject = self._test_document.file_latest.get_file_metadata(
            dotted_name=TEST_MSG_FILE_METADATA_DOTTED_NAME_SUBJECT
        )
        self.assertEqual(value_subject, TEST_MSG_FILE_METADATA_VALUE_SUBJECT)

        value_to = self._test_document.file_latest.get_file_metadata(
            dotted_name=TEST_MSG_FILE_METADATA_DOTTED_NAME_TO
        )
        self.assertEqual(value_to, TEST_MSG_FILE_METADATA_VALUE_TO)



from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0088_populate_version_active'),
        ('sources', '0029_update_source_backend_paths')
    ]

    operations = [
        migrations.CreateModel(
            name='DocumentFileSourceMetadata',
            fields=[
                (
                    'id', models.AutoField(
                        auto_created=True, primary_key=True,
                        serialize=False, verbose_name='ID'
                    )
                ),
                (
                    'key', models.CharField(
                        db_index=True, help_text='Name of the source '
                        'metadata entry.', max_length=255, verbose_name='Key'
                    )
                ),
                (
                    'value', models.TextField(
                        blank=True, help_text='The actual value '
                        'stored in the source metadata for the document.',
                        null=True, verbose_name='Value'
                    )
                ),
                (
                    'document_file', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='source_metadata',
                        to='documents.documentfile',
                        verbose_name='Document file')
                ),
                (
                    'source', models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='metadata', to='sources.source',
                        verbose_name='Source'
                    )
                )
            ],
            options={
                'verbose_name': 'Document file source metadata',
                'verbose_name_plural': 'Document file source metadata',
                'ordering': ('key',),
                'unique_together': {
                    ('source', 'document_file', 'key')
                }
            }
        )
    ]



import json

from mayan.apps.documents.tests.base import DocumentTestMixin
from mayan.apps.source_web_forms.tests.mixins import WebFormSourceTestMixin
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceActionViewTestMixin
from mayan.apps.testing.tests.mixins import TestMixinObjectCreationTrack

from ..models import Quota

from .literals import (
    TEST_QUOTA_DATA, TEST_QUOTA_DOTTED_PATH, TEST_QUOTA_TEST_LIMIT_EDITED,
    TEST_QUOTA_WITH_MIXINS_DOTTED_PATH
)


class QuotaTestMixin(TestMixinObjectCreationTrack):
    _test_object_model = Quota
    _test_object_name = '_test_quota'

    def _create_test_quota(self):
        self._test_quota = Quota.objects.create(
            backend_data=json.dumps(obj=TEST_QUOTA_DATA),
            backend_path=TEST_QUOTA_DOTTED_PATH
        )

    def _create_test_quota_with_mixins(self):
        self._test_quota = Quota.objects.create(
            backend_data=json.dumps(obj=TEST_QUOTA_DATA),
            backend_path=TEST_QUOTA_WITH_MIXINS_DOTTED_PATH
        )


class DocumentViewQuotaHookTestMixin(
    DocumentTestMixin, SourceActionViewTestMixin,
    WebFormSourceTestMixin
):
    """
    Combined class for source document upload view quota testing.
    """


class QuotaViewTestMixin(QuotaTestMixin):
    def _request_test_quota_backend_selection_get_view(self):
        return self.get(viewname='quotas:quota_backend_selection')

    def _request_test_quota_create_get_view(self):
        self._test_object_track()

        response = self.get(
            viewname='quotas:quota_create', kwargs={
                'class_path': TEST_QUOTA_DOTTED_PATH
            }
        )

        self._test_object_set()

        return response

    def _request_test_quota_with_mixins_create_get_view(self):
        self._test_object_track()

        response = self.get(
            viewname='quotas:quota_create', kwargs={
                'class_path': TEST_QUOTA_WITH_MIXINS_DOTTED_PATH
            }
        )

        self._test_object_set()

        return response

    def _request_test_quota_create_post_view(self):
        self._test_object_track()

        response = self.post(
            viewname='quotas:quota_create', kwargs={
                'class_path': TEST_QUOTA_DOTTED_PATH
            }, data=TEST_QUOTA_DATA
        )

        self._test_object_set()

        return response

    def _request_test_quota_delete_view(self):
        return self.post(
            viewname='quotas:quota_delete', kwargs={
                'quota_id': self._test_quota.pk
            }
        )

    def _request_test_quota_edit_view(self):
        return self.post(
            viewname='quotas:quota_edit', kwargs={
                'quota_id': self._test_quota.pk
            }, data={
                'test_limit': TEST_QUOTA_TEST_LIMIT_EDITED
            }
        )

    def _request_test_quota_list_view(self):
        return self.get(viewname='quotas:quota_list')



from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.classes import PythonDependency
from mayan.apps.dependencies.environments import (
    environment_build, environment_development, environment_documentation,
    environment_documentation_override
)
from mayan.settings.literals import PYTHON_WHEEL_VERSION

PythonDependency(
    legal_text='''
        Copyright (c) Django Software Foundation and individual contributors.
        All rights reserved.

        Redistribution and use in source and binary forms, with or without modification,
        are permitted provided that the following conditions are met:

        1. Redistributions of source code must retain the above copyright notice,
        this list of conditions and the following disclaimer.

        2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

        3. Neither the name of Django nor the names of its contributors may be used
        to endorse or promote products derived from this software without
        specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
        ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
        WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
        DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
        ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
        (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
        LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
        ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    ''', module=__name__, name='django', version_string='==3.2.25'
)
PythonDependency(
    legal_text='''
        Copyright (c) 2006 Kirill Simonov

        Permission is hereby granted, free of charge, to any person obtaining a copy of
        this software and associated documentation files (the "Software"), to deal in
        the Software without restriction, including without limitation the rights to
        use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
        of the Software, and to permit persons to whom the Software is furnished to do
        so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.
    ''', module=__name__, name='PyYAML', version_string='==6.0.1'
)
PythonDependency(
    legal_text='''
        Copyright (c) 2009-2015, Carl Meyer and contributors
        All rights reserved.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are
        met:

        * Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above
        copyright notice, this list of conditions and the following
        disclaimer in the documentation and/or other materials provided
        with the distribution.
        * Neither the name of the author nor the names of other
        contributors may be used to endorse or promote products derived
        from this software without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
        "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
        LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
        A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
        OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
        SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
        LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
        DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
        THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    ''', module=__name__, name='django-model-utils', version_string='==4.3.1'
)
PythonDependency(
    legal_text='''
        Django MPTT
        -----------

        Copyright (c) 2007, Jonathan Buchanan

        Permission is hereby granted, free of charge, to any person obtaining a copy of
        this software and associated documentation files (the "Software"), to deal in
        the Software without restriction, including without limitation the rights to
        use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
        the Software, and to permit persons to whom the Software is furnished to do so,
        subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
        FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
        COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
        IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
        CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
    ''', module=__name__, name='django-mptt', version_string='==0.14.0'
)
PythonDependency(
    module=__name__, name='importlib-metadata', version_string='==6.8.0'
)
PythonDependency(
    legal_text='''
        Author: Christian Theune
        License: LGPL 2.1
    ''', module=__name__, name='pycountry', version_string='==22.3.5'
)
PythonDependency(
    module=__name__, name='requests', version_string='==2.31.0'
)
PythonDependency(
    module=__name__, name='setuptools', version_string='==69.1.1'
)
PythonDependency(
    legal_text='''
        Copyright (C) 2011-2012 by Andrew Moffat

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in
        all copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
        THE SOFTWARE.
    ''', module=__name__, name='sh', version_string='==2.0.4'
)

# Build

PythonDependency(
    environment=environment_build, module=__name__, name='twine',
    version_string='==4.0.2'
)
PythonDependency(
    environments=(environment_build, environment_documentation),
    module=__name__, name='wheel', version_string='=={}'.format(
        PYTHON_WHEEL_VERSION
    )
)

# Development

PythonDependency(
    module=__name__, environment=environment_development,
    name='devpi-server', version_string='==6.5.0'
)
PythonDependency(
    environment=environment_development, module=__name__,
    name='django-debug-toolbar', version_string='==3.2.4'
)
PythonDependency(
    environment=environment_development, module=__name__,
    name='django-extensions', version_string='==3.2.3'
)
PythonDependency(
    environment=environment_development, help_text=_(
        'Used to allow offline translation of the code text strings.'
    ), module=__name__, name='django-rosetta', version_string='==0.9.9'
)
PythonDependency(
    environment=environment_development, module=__name__,
    name='django-silk', version_string='==5.0.3'
)
PythonDependency(
    environment=environment_development, help_text=_(
        'Provides style checking.'
    ), module=__name__, name='flake8', version_string='==7.0.0'
)
PythonDependency(
    environment=environment_development, help_text=_(
        'Command line environment with autocompletion.'
    ), module=__name__, name='ipython', version_string='==8.21.0'
)
PythonDependency(
    module=__name__, environment=environment_development,
    name='pdbpp', version_string='==0.10.3'
)
PythonDependency(
    environment=environment_development, help_text=_(
        'Checks proper formatting of the README file.'
    ), module=__name__, name='readme', version_string='==0.7.1'
)
PythonDependency(
    environment=environment_development,
    module=__name__, name='safety', version_string='==3.0.1'
)

# Documentation

PythonDependency(
    environment=environment_documentation, module=__name__, name='Sphinx',
    version_string='==5.3.0'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinx-autobuild', version_string='==2021.3.14'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinx-sitemap', version_string='==2.2.0'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinx_rtd_theme', version_string='==2.0.0'
)
PythonDependency(
    environment=environment_documentation, module=__name__,
    name='sphinxcontrib-spelling', version_string='==7.1.0'
)
PythonDependency(
    environment=environment_documentation, module=__name__, name='tornado',
    version_string='==6.1'
)
PythonDependency(
    environment=environment_documentation_override, module=__name__,
    name='jinja2', version_string='==3.0.3'
)



from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.source_compressed.source_backends.literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK,
    SOURCE_UNCOMPRESS_CHOICE_NEVER
)
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceActionViewTestMixin

from .mixins import CompressedSourceTestMixin


class CompressedSourceBackendActionDocumentUploadViewTestCase(
    CompressedSourceTestMixin, SourceActionViewTestMixin,
    GenericDocumentViewTestCase
):
    _test_source_create_auto = False
    auto_upload_test_document = False

    def test_always(self):
        self._test_source_create(
            extra_data={
                'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS
            }
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_document_upload_get_view()
        self.assertNotContains(
            response=response, status_code=200, text='source-expand'
        )

        self.assertEqual(
            Document.objects.count(), test_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_ask(self):
        self._test_source_create(
            extra_data={
                'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK
            }
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_document_upload_get_view()
        self.assertContains(
            response=response, status_code=200, text='source-expand'
        )

        self.assertEqual(
            Document.objects.count(), test_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_never(self):
        self._test_source_create(
            extra_data={
                'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER
            }
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        test_document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_document_upload_get_view()
        self.assertNotContains(
            response=response, status_code=200, text='source-expand'
        )

        self.assertEqual(
            Document.objects.count(), test_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import logging

from django.template import RequestContext
from django.urls import reverse
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions.models import Role
from mayan.apps.views.generics import (
    AddRemoveView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectListView
)
from mayan.apps.views.view_mixins import (
    ContentTypeViewMixin, ExternalObjectViewMixin
)

from .classes import ModelPermission
from .forms import ACLCreateForm
from .icons import (
    icon_acl_create, icon_acl_delete, icon_acl_list, icon_acl_permissions,
    icon_global_acl_list
)
from .links import link_acl_create
from .models import AccessControlList, GlobalAccessControlListProxy
from .permissions import permission_acl_edit, permission_acl_view

logger = logging.getLogger(name=__name__)


class ACLCreateView(
    ContentTypeViewMixin, ExternalObjectViewMixin, SingleObjectCreateView
):
    content_type_url_kw_args = {
        'app_label': 'app_label',
        'model_name': 'model_name'
    }
    external_object_permission = permission_acl_edit
    external_object_pk_url_kwarg = 'object_id'
    form_class = ACLCreateForm
    view_icon = icon_acl_create

    def get_error_message_duplicate(self):
        return _(
            'An ACL for "%(object)s" using role "%(role)s" already exists. '
            'Edit that ACL entry instead.'
        ) % {
            'object': self.get_external_object(), 'role': self.object.role
        }

    def get_external_object_queryset(self):
        # Here we get a queryset the object model for which an ACL will be
        # created.
        return self.get_content_type().get_all_objects_for_this_type()

    def get_extra_context(self):
        return {
            'object': self.external_object,
            'title': _(
                'New access control lists for: %s'
            ) % self.external_object
        }

    def get_form_extra_kwargs(self):
        try:
            roles = self.external_object.acls.values('role')
        except AttributeError:
            # Fallback when attempting to access the ACLs generic relation
            # field of models that have not been registered.
            roles = Role.objects.none()

        return {
            'field_name': 'role',
            'label': _('Role'),
            'queryset': Role.objects.exclude(pk__in=roles),
            'widget_attributes': {'class': 'select2'},
            'user': self.request.user
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'content_object': self.external_object
        }

    def get_queryset(self):
        self.external_object.acls.all()

    def get_success_url(self):
        return self.object.get_absolute_url()


class ACLDeleteView(SingleObjectDeleteView):
    model = AccessControlList
    object_permission = permission_acl_edit
    pk_url_kwarg = 'acl_id'
    view_icon = icon_acl_delete

    def get_extra_context(self):
        return {
            'acl': self.object,
            'navigation_object_list': ('object', 'acl'),
            'object': self.object.content_object,
            'title': _('Delete ACL: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_post_action_redirect(self):
        return reverse(
            viewname='acls:acl_list', kwargs={
                'app_label': self.object.content_type.app_label,
                'model_name': self.object.content_type.model,
                'object_id': self.object.object_id
            }
        )


class ACLListView(
    ContentTypeViewMixin, ExternalObjectViewMixin, SingleObjectListView
):
    content_type_url_kw_args = {
        'app_label': 'app_label',
        'model_name': 'model_name'
    }
    external_object_permission = permission_acl_view
    external_object_pk_url_kwarg = 'object_id'
    view_icon = icon_acl_list

    def get_external_object_queryset(self):
        # Here we get a queryset the object model for which an ACL will be
        # created.
        return self.get_content_type().get_all_objects_for_this_type()

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_acl_list,
            'no_results_main_link': link_acl_create.resolve(
                context=RequestContext(
                    self.request, {
                        'resolved_object': self.external_object
                    }
                )
            ),
            'no_results_title': _(
                'There are no ACLs for this object'
            ),
            'no_results_text': _(
                'ACL stands for Access Control List and is a precise method '
                'to control user access to objects in the system. ACLs '
                'allow granting a permission to a role but only for a '
                'specific object or set of objects.'
            ),
            'object': self.external_object,
            'title': _(
                'Access control lists for: %s' % self.external_object
            )
        }

    def get_source_queryset(self):
        return self.external_object.acls.all()


class ACLPermissionAddRemoveView(AddRemoveView):
    list_added_title = _('Granted permissions')
    list_available_title = _('Available permissions')
    main_object_method_add_name = 'permissions_add'
    main_object_method_remove_name = 'permissions_remove'
    main_object_model = AccessControlList
    main_object_permission = permission_acl_edit
    main_object_pk_url_kwarg = 'acl_id'
    related_field = 'permissions'
    view_icon = icon_acl_permissions

    def generate_choices(self, queryset):
        namespaces_dictionary = {}

        # Sort permissions by their translatable label
        object_list = sorted(
            queryset,
            key=lambda permission: permission.volatile_permission.label
        )

        # Group permissions by namespace
        for permission in object_list:
            namespaces_dictionary.setdefault(
                permission.volatile_permission.namespace.label, []
            )
            namespaces_dictionary[
                permission.volatile_permission.namespace.label
            ].append(
                (
                    permission.pk, force_text(s=permission)
                )
            )

        # Sort permissions by their translatable namespace label.
        return sorted(
            namespaces_dictionary.items()
        )

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_disabled_choices(self):
        """
        Get permissions from a parent's ACLs or directly granted to the role.
        We return a list since that is what the form widget's can process.
        """
        return self.main_object.get_inherited_permissions().values_list(
            'pk', flat=True
        )

    def get_extra_context(self):
        return {
            'acl': self.main_object,
            'object': self.main_object.content_object,
            'navigation_object_list': ('object', 'acl'),
            'title': _('Role "%(role)s" permission\'s for "%(object)s".') % {
                'object': self.main_object.content_object,
                'role': self.main_object.role
            }
        }

    def get_list_added_help_text(self):
        if self.main_object.get_inherited_permissions():
            return _(
                'Disabled permissions are inherited from a parent object or '
                'directly granted to the role and can\'t be removed from '
                'this view. Inherited permissions need to be removed from '
                'the parent object\'s ACL or from them role via the '
                'Setup menu.'
            )
        else:
            return super().get_list_added_help_text()

    def get_list_added_queryset(self):
        """
        Merge of permissions we hold for this object and the permissions we
        hold for this object's parents via another ACL. .distinct() is added
        in case the permission was added to the ACL and then added to a
        parent ACL's and thus inherited and would appear twice. If
        order to remove the double permission from the ACL it would need to be
        remove from the parent first to enable the choice in the form,
        remove it from the ACL and then re-add it to the parent ACL.
        """
        queryset_acl = super().get_list_added_queryset()

        return (
            queryset_acl | self.main_object.get_inherited_permissions()
        ).distinct()

    def get_secondary_object_source_queryset(self):
        return ModelPermission.get_for_instance(
            instance=self.main_object.content_object
        )


class GlobalACLListView(SingleObjectListView):
    model = GlobalAccessControlListProxy
    object_permission = permission_acl_view
    view_icon = icon_global_acl_list

    def get_extra_context(self):
        return {
            'hide_object': True,
            'no_results_icon': icon_acl_list,
            'no_results_title': _(
                'There are no ACLs'
            ),
            'no_results_text': _(
                'ACL stands for Access Control List and is a precise method '
                'to control user access to objects in the system. ACLs '
                'allow granting a permission to a role but only for a '
                'specific object or set of objects.'
            ),
            'title': _(
                'Global access control lists'
            )
        }



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_version_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.sources.permissions import permission_sources_edit
from mayan.apps.sources.tests.mixins.source_view_mixins import SourceViewTestMixin

from .mixins import PeriodicSourceBackendTestMixin


class PeriodicSourceViewTestCase(
    PeriodicSourceBackendTestMixin, SourceViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_source_test_get_view_no_permission(self):
        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_get_view_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_disabled_get_view_with_access(self):
        self._test_source.enabled = False
        self._test_source.save()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_get_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_no_permission(self):
        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            Document.objects.count(), document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_source_test_post_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

    def test_source_test_disabled_post_view_with_access(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source.enabled = False
        self._test_source.save()

        self.grant_access(
            obj=self._test_source, permission=permission_sources_edit
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_source_test_post_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0020_auto_20150714_2233'),
    ]

    operations = [
        migrations.AlterField(
            model_name='documentversion',
            name='comment',
            field=models.TextField(
                default='', null=True, verbose_name='Comment', blank=True
            ),
            preserve_default=True,
        ),
    ]



from mayan.apps.testing.tests.base import MayanMigratorTestCase


class WorkflowTemplateTransitionTriggerMigrationTestCase(
    MayanMigratorTestCase
):
    migrate_from = ('document_states', '0023_auto_20200930_0726')
    migrate_to = (
        'document_states',
        '0024_remove_duplicate_workflow_template_transition_triggers'
    )

    def prepare(self):
        StoredEventType = self.old_state.apps.get_model(
            app_label='events', model_name='StoredEventType'
        )

        Workflow = self.old_state.apps.get_model(
            app_label='document_states', model_name='Workflow'
        )
        WorkflowState = self.old_state.apps.get_model(
            app_label='document_states', model_name='WorkflowState'
        )
        WorkflowTransition = self.old_state.apps.get_model(
            app_label='document_states', model_name='WorkflowTransition'
        )

        WorkflowTransitionTriggerEvent = self.old_state.apps.get_model(
            app_label='document_states',
            model_name='WorkflowTransitionTriggerEvent'
        )

        test_stored_event_type = StoredEventType.objects.create(
            name='test'
        )
        test_workflow_template = Workflow.objects.create(label='test')
        test_workflow_state_0 = WorkflowState.objects.create(
            label='test workflow template state 0',
            workflow=test_workflow_template
        )
        test_workflow_state_1 = WorkflowState.objects.create(
            label='test workflow template state 1',
            workflow=test_workflow_template
        )
        test_workflow_template_transition = WorkflowTransition.objects.create(
            origin_state=test_workflow_state_0,
            destination_state=test_workflow_state_1,
            workflow=test_workflow_template
        )

        WorkflowTransitionTriggerEvent.objects.create(
            event_type=test_stored_event_type,
            transition_id=test_workflow_template_transition.pk
        )
        WorkflowTransitionTriggerEvent.objects.create(
            event_type=test_stored_event_type,
            transition_id=test_workflow_template_transition.pk
        )
        self.assertTrue(
            WorkflowTransitionTriggerEvent.objects.count(), 2
        )

    def test_duplicated_workflow_template_transition_trigger_removal(self):
        WorkflowTransitionTriggerEvent = self.new_state.apps.get_model(
            app_label='document_states',
            model_name='WorkflowTransitionTriggerEvent'
        )

        self.assertTrue(
            WorkflowTransitionTriggerEvent.objects.count(), 1
        )



from django import forms
from django.forms.formsets import formset_factory
from django.utils.translation import ugettext_lazy as _

from .models import EventSubscription, ObjectEventSubscription


class EventTypeUserRelationshipForm(forms.Form):
    namespace = forms.CharField(
        label=_('Namespace'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    label = forms.CharField(
        label=_('Label'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    subscription = forms.ChoiceField(
        label=_('Subscription'),
        widget=forms.RadioSelect(), choices=(
            ('none', _('No')),
            ('subscribed', _('Subscribed'))
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, **kwargs
        )

        if 'stored_event_type' in self.initial:
            self.fields['namespace'].initial = self.initial['stored_event_type'].namespace
            self.fields['label'].initial = self.initial['stored_event_type'].label

            subscription = EventSubscription.objects.get_for(
                stored_event_type=self.initial['stored_event_type'],
                user=self.initial['user'],
            )

            if subscription.exists():
                self.fields['subscription'].initial = 'subscribed'
            else:
                self.fields['subscription'].initial = 'none'

    def save(self):
        subscription = EventSubscription.objects.get_for(
            stored_event_type=self.initial['stored_event_type'],
            user=self.initial['user'],
        )

        if self.cleaned_data['subscription'] == 'none':
            subscription.delete()
        elif self.cleaned_data['subscription'] == 'subscribed':
            if not subscription.exists():
                EventSubscription.objects.create_for(
                    stored_event_type=self.initial['stored_event_type'],
                    user=self.initial['user']
                )


EventTypeUserRelationshipFormSet = formset_factory(
    form=EventTypeUserRelationshipForm, extra=0
)


class ObjectEventTypeUserRelationshipForm(forms.Form):
    namespace = forms.CharField(
        label=_('Namespace'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    label = forms.CharField(
        label=_('Label'), required=False,
        widget=forms.TextInput(
            attrs={'readonly': 'readonly'}
        )
    )
    subscription = forms.ChoiceField(
        label=_('Subscription'),
        widget=forms.RadioSelect(), choices=(
            ('none', _('No')),
            ('subscribed', _('Subscribed'))
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args, **kwargs
        )

        self.fields['namespace'].initial = self.initial['stored_event_type'].namespace
        self.fields['label'].initial = self.initial['stored_event_type'].label

        subscription = ObjectEventSubscription.objects.get_for(
            obj=self.initial['object'],
            stored_event_type=self.initial['stored_event_type'],
            user=self.initial['user'],
        )

        if subscription.exists():
            self.fields['subscription'].initial = 'subscribed'
        else:
            self.fields['subscription'].initial = 'none'

    def save(self):
        subscription = ObjectEventSubscription.objects.get_for(
            obj=self.initial['object'],
            stored_event_type=self.initial['stored_event_type'],
            user=self.initial['user'],
        )

        if self.cleaned_data['subscription'] == 'none':
            subscription.delete()
        elif self.cleaned_data['subscription'] == 'subscribed':
            if not subscription.exists():
                ObjectEventSubscription.objects.create_for(
                    obj=self.initial['object'],
                    stored_event_type=self.initial['stored_event_type'],
                    user=self.initial['user']
                )


ObjectEventTypeUserRelationshipFormSet = formset_factory(
    form=ObjectEventTypeUserRelationshipForm, extra=0
)



from django.db import migrations, models

import mayan.apps.common.validators
import mayan.apps.documents.classes
import mayan.apps.documents.models.document_file_models
import mayan.apps.storage.classes


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0054_trasheddocument')
    ]

    operations = [
        migrations.AddField(
            model_name='documenttype',
            name='filename_generator_backend',
            field=models.CharField(
                default='uuid',
                help_text='The class responsible for producing the actual '
                'filename used to store the uploaded documents.',
                max_length=224, verbose_name='Filename generator backend'
            )
        ),
        migrations.AddField(
            model_name='documenttype',
            name='filename_generator_backend_arguments',
            field=models.TextField(
                blank=True, help_text='The arguments for the filename '
                'generator backend as a YAML dictionary.', validators=[
                    mayan.apps.common.validators.YAMLValidator()
                ], verbose_name='Filename generator backend arguments'
            )
        ),
        migrations.AlterField(
            model_name='documentversion',
            name='file',
            field=models.FileField(
                storage=mayan.apps.storage.classes.DefinedStorageLazy(
                    name='documents__documentversion'
                ), upload_to=mayan.apps.documents.models.document_file_models.upload_to,
                verbose_name='File'
            )
        )
    ]



from ..interfaces import (
    SourceBackendActionInterface, SourceBackendActionInterfaceRequestRESTAPI,
    SourceBackendActionInterfaceRequestViewForm,
    SourceBackendActionInterfaceTask
)

from .arguments import argument_document_file_filename


class SourceBackendActionMixinDocumentFileFilenameInteractive:
    class Interface:
        class Model(SourceBackendActionInterface):
            class Argument:
                filename = argument_document_file_filename

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['filename'] = self.context['filename']

        class RESTAPI(SourceBackendActionInterfaceRequestRESTAPI):
            class Argument:
                filename = argument_document_file_filename

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['filename'] = self.context['filename']

        class Task(SourceBackendActionInterfaceTask):
            class Argument:
                filename = argument_document_file_filename

            def process_interface_context(self):
                super().process_interface_context()

                self.action_kwargs['filename'] = self.context['filename']

        class View(SourceBackendActionInterfaceRequestViewForm):
            def process_interface_context(self):
                super().process_interface_context()

                form_document = self.context['forms']['document_form']
                form_document_cleaned_data = form_document.cleaned_data

                self.action_kwargs['filename'] = form_document_cleaned_data.get(
                    'filename', None
                )

    def get_document_file_task_kwargs(self, filename=None, **kwargs):
        result = super().get_document_file_task_kwargs(**kwargs)

        result['filename'] = filename

        return result

    def get_task_kwargs(self, filename=None, **kwargs):
        result = super().get_task_kwargs(**kwargs)

        result['action_interface_kwargs']['filename'] = filename

        return result



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_SOURCES_BACKEND_ARGUMENTS, DEFAULT_SOURCES_CACHE_STORAGE_BACKEND,
    DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS
)
from .setting_migrations import SourcesSettingMigration

namespace = SettingNamespace(
    label=_('Sources'), migration_class=SourcesSettingMigration,
    name='sources', version='0003'
)

setting_backend_arguments = namespace.add_setting(
    default=DEFAULT_SOURCES_BACKEND_ARGUMENTS,
    global_name='SOURCES_BACKEND_ARGUMENTS', help_text=_(
        'Arguments to use when creating source backends.'
    )
)
setting_source_cache_storage_backend = namespace.add_setting(
    global_name='SOURCES_CACHE_STORAGE_BACKEND',
    default=DEFAULT_SOURCES_CACHE_STORAGE_BACKEND, help_text=_(
        'Path to the Storage subclass used to store cached '
        'source image files.'
    )
)
setting_source_cache_storage_backend_arguments = namespace.add_setting(
    global_name='SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS',
    default=DEFAULT_SOURCES_CACHE_STORAGE_BACKEND_ARGUMENTS,
    help_text=_(
        'Arguments to pass to the SOURCES_SOURCE_CACHE_STORAGE_BACKEND.'
    )
)



import json

from mayan.apps.document_states.literals import WORKFLOW_ACTION_ON_ENTRY
from mayan.apps.document_states.tests.mixins.workflow_template_state_action_mixins import WorkflowTemplateStateActionTestMixin
from mayan.apps.testing.tests.base import BaseTestCase, GenericViewTestCase

from ..models.trashed_document_models import TrashedDocument
from ..workflow_actions import DocumentTypeChangeAction, TrashDocumentAction

from .literals import (
    TEST_DOCUMENT_TYPE_CHANGE_ACTION_DOTTED_PATH,
    TEST_TRASH_DOCUMENT_WORKFLOW_ACTION_DOTTED_PATH
)


class WorkflowTemplateStateActionTestCase(
    WorkflowTemplateStateActionTestMixin, BaseTestCase
):
    def test_document_type_change_action(self):
        document_type = self._test_document.document_type

        self._create_test_document_type()

        self._execute_workflow_template_state_action(
            klass=DocumentTypeChangeAction, kwargs={
                'document_type': self._test_document_types[1].pk
            }
        )

        self.assertNotEqual(
            self._test_document.document_type, document_type
        )

    def test_trash_document_action(self):
        trashed_document_count = TrashedDocument.objects.count()

        self._execute_workflow_template_state_action(
            klass=TrashDocumentAction
        )
        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count + 1
        )


class WorkflowTemplateStateActionTransitionTestCase(
    WorkflowTemplateStateActionTestMixin, GenericViewTestCase
):
    def test_document_type_change_action_execution(self):
        self._create_test_document_type()

        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._test_workflow_template_states[1].actions.create(
            backend_data=json.dumps(
                obj={'document_type': self._test_document_types[1].pk}
            ),
            backend_path=TEST_DOCUMENT_TYPE_CHANGE_ACTION_DOTTED_PATH,
            label='', when=WORKFLOW_ACTION_ON_ENTRY,

        )
        self._test_workflow_template.document_types.add(
            self._test_document_types[0]
        )

        document_type = self._test_document.document_type

        self._test_document.workflows.first().do_transition(
            transition=self._test_workflow_template_transition
        )

        self.assertNotEqual(
            self._test_document.document_type, document_type
        )

    def test_trash_document_action_workflow_execution(self):
        self._create_test_workflow_template_state()
        self._create_test_workflow_template_transition()
        self._test_workflow_template_states[1].actions.create(
            backend_path=TEST_TRASH_DOCUMENT_WORKFLOW_ACTION_DOTTED_PATH,
            label='', when=WORKFLOW_ACTION_ON_ENTRY,
        )
        self._test_workflow_template.document_types.add(
            self._test_document_type
        )

        trashed_document_count = TrashedDocument.objects.count()

        self._test_document.workflows.first().do_transition(
            transition=self._test_workflow_template_transition
        )

        self.assertEqual(
            TrashedDocument.objects.count(), trashed_document_count + 1
        )



from django.apps import apps
from django.contrib.auth import get_user_model

from mayan.celery import app


@app.task(ignore_result=True)
def task_send_object(
    content_type_id, body, object_id, sender, subject, recipient,
    user_mailer_id, as_attachment=False,
    content_function_dotted_path=None,
    mime_type_function_dotted_path=None, object_name=None,
    organization_installation_url=None, user_id=None
):
    ContentType = apps.get_model(
        app_label='contenttypes', model_name='ContentType'
    )
    UserMailer = apps.get_model(
        app_label='mailer', model_name='UserMailer'
    )
    User = get_user_model()

    content_type = ContentType.objects.get(pk=content_type_id)
    obj = content_type.get_object_for_this_type(pk=object_id)

    user_mailer = UserMailer.objects.get(pk=user_mailer_id)

    if user_id:
        user = User.objects.get(pk=user_id)
    else:
        user = None

    user_mailer.send_object(
        as_attachment=as_attachment, body=body,
        content_function_dotted_path=content_function_dotted_path,
        mime_type_function_dotted_path=mime_type_function_dotted_path,
        obj=obj, object_name=object_name,
        organization_installation_url=organization_installation_url,
        subject=subject, to=recipient, user=user
    )



from mayan.apps.appearance.classes import Icon

# Document version

icon_document_version_ocr_content_single_delete = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='times'
)
icon_document_version_ocr_content_multiple_delete = icon_document_version_ocr_content_single_delete
icon_document_version_ocr_content_detail = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_single_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)
icon_document_version_ocr_multiple_submit = icon_document_version_ocr_single_submit

# Document version page

icon_document_version_page_ocr_content_detail = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_page_ocr_content_edit = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_version_ocr_content_download = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-down'
)

# Document type

icon_document_type_ocr_settings = Icon(
    driver_name='fontawesome', symbol='font'
)
icon_document_type_ocr_submit = Icon(
    driver_name='fontawesome-dual', primary_symbol='font',
    secondary_symbol='arrow-right'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Storage'), name='storage'
)

permission_download_file_delete = namespace.add_permission(
    label=_('Delete user files'), name='download_file_delete'
)
permission_download_file_download = namespace.add_permission(
    label=_('Download user files'), name='download_file_download'
)
permission_download_file_view = namespace.add_permission(
    label=_('View user files'), name='download_file_view'
)



from .literals import TEST_RECEIVE_KEY, TEST_SEARCH_FINGERPRINT


def mock_recv_keys(self, keyserver, *keyids):
    class ImportResult:
        count = 1
        fingerprints = [TEST_SEARCH_FINGERPRINT]

    self.import_keys(TEST_RECEIVE_KEY)

    return ImportResult()



import logging

from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _
from django.views.generic import RedirectView

from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.views.generics import (
    AddRemoveView, SingleObjectCreateView, SingleObjectDeleteView,
    SingleObjectEditView, SingleObjectListView
)
from mayan.apps.views.view_mixins import ExternalObjectViewMixin

from .forms import WebLinkForm
from .icons import icon_web_link_setup
from .links import link_web_link_create
from .models import ResolvedWebLink, WebLink
from .permissions import (
    permission_web_link_create, permission_web_link_delete,
    permission_web_link_edit, permission_web_link_instance_view,
    permission_web_link_view
)

logger = logging.getLogger(name=__name__)


class DocumentTypeWebLinksView(AddRemoveView):
    list_added_title = _('Web links enabled')
    list_available_title = _('Available web links')
    main_object_method_add_name = 'web_links_add'
    main_object_method_remove_name = 'web_links_remove'
    main_object_model = DocumentType
    main_object_permission = permission_document_type_edit
    main_object_pk_url_kwarg = 'document_type_id'
    related_field = 'web_links'
    secondary_object_model = WebLink
    secondary_object_permission = permission_web_link_edit

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(
                'Web links to enable for document type: %s'
            ) % self.main_object
        }


class ResolvedWebLinkView(ExternalObjectViewMixin, RedirectView):
    external_object_pk_url_kwarg = 'document_id'
    external_object_permission = permission_web_link_instance_view
    external_object_queryset = Document.valid.all()

    def get_redirect_url(self, *args, **kwargs):
        return self.get_web_link().get_redirect(
            document=self.external_object, user=self.request.user
        ).url

    def get_web_link(self):
        return get_object_or_404(
            klass=self.get_web_link_queryset(), pk=self.kwargs['web_link_id']
        )

    def get_web_link_queryset(self):
        queryset = ResolvedWebLink.objects.get_for(
            document=self.external_object, user=self.request.user
        )
        return AccessControlList.objects.restrict_queryset(
            permission=permission_web_link_instance_view, queryset=queryset,
            user=self.request.user
        )


class WebLinkCreateView(SingleObjectCreateView):
    extra_context = {
        'title': _('Create new web link')
    }
    form_class = WebLinkForm
    post_action_redirect = reverse_lazy(
        viewname='web_links:web_link_list'
    )
    view_permission = permission_web_link_create

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class WebLinkDeleteView(SingleObjectDeleteView):
    model = WebLink
    object_permission = permission_web_link_delete
    pk_url_kwarg = 'web_link_id'
    post_action_redirect = reverse_lazy(
        viewname='web_links:web_link_list'
    )

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Delete web link: %s') % self.object
        }


class WebLinkDocumentTypesView(AddRemoveView):
    list_added_title = _('Document types enabled')
    list_available_title = _('Available document types')
    main_object_method_add_name = 'document_types_add'
    main_object_method_remove_name = 'document_types_remove'
    main_object_model = WebLink
    main_object_permission = permission_web_link_edit
    main_object_pk_url_kwarg = 'web_link_id'
    related_field = 'document_types'
    secondary_object_model = DocumentType
    secondary_object_permission = permission_document_type_edit

    def get_actions_extra_kwargs(self):
        return {'user': self.request.user}

    def get_extra_context(self):
        return {
            'object': self.main_object,
            'title': _(
                'Document type for which to enable web link: %s'
            ) % self.main_object
        }


class WebLinkEditView(SingleObjectEditView):
    form_class = WebLinkForm
    model = WebLink
    object_permission = permission_web_link_edit
    pk_url_kwarg = 'web_link_id'
    post_action_redirect = reverse_lazy(
        viewname='web_links:web_link_list'
    )

    def get_extra_context(self):
        return {
            'object': self.object,
            'title': _('Edit web link: %s') % self.object
        }

    def get_instance_extra_data(self):
        return {'_event_actor': self.request.user}


class WebLinkListView(SingleObjectListView):
    object_permission = permission_web_link_view

    def get_extra_context(self):
        return {
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_web_link_setup,
            'no_results_main_link': link_web_link_create.resolve(
                context=RequestContext(request=self.request)
            ),
            'no_results_text': _(
                'Web links allow generating HTTP links from documents to '
                'external resources. The link URL\'s can contain document '
                'properties values.'
            ),
            'no_results_title': _(
                'There are no web links'
            ),
            'title': _('Web links')
        }

    def get_source_queryset(self):
        return self.get_web_link_queryset()

    def get_web_link_queryset(self):
        return WebLink.objects.all()


class DocumentWebLinkListView(ExternalObjectViewMixin, WebLinkListView):
    external_object_permission = permission_web_link_instance_view
    external_object_pk_url_kwarg = 'document_id'
    external_object_queryset = Document.valid.all()
    object_permission = permission_web_link_instance_view

    def get_extra_context(self):
        return {
            'document': self.external_object,
            'hide_link': True,
            'hide_object': True,
            'no_results_icon': icon_web_link_setup,
            'no_results_text': _(
                'Web links allow generating HTTP links from documents to '
                'external resources. The link URL\'s can contain document '
                'properties values.'
            ),
            'no_results_title': _(
                'There are no web links for this document'
            ),
            'object': self.external_object,
            'title': _('Web links for document: %s') % self.external_object
        }

    def get_web_link_queryset(self):
        return ResolvedWebLink.objects.get_for(
            document=self.external_object, user=self.request.user
        )



from django.conf import settings
from django.contrib.auth import get_user_model
from django.db.models.signals import post_save
from django.utils.translation import ugettext_lazy as _

from mayan.apps.acls.classes import ModelPermission
from mayan.apps.acls.permissions import (
    permission_acl_edit, permission_acl_view
)
from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.classes import ModelCopy
from mayan.apps.common.menus import (
    menu_list_facet, menu_object, menu_return, menu_secondary, menu_setup
)
from mayan.apps.events.classes import EventModelRegistry, ModelEventType
from mayan.apps.navigation.classes import SourceColumn

from .events import event_theme_edited, event_user_theme_settings_edited
from .handlers import handler_user_theme_setting_create
from .links import (
    link_theme_create, link_theme_delete, link_theme_edit, link_theme_list,
    link_theme_setup, link_user_theme_settings_detail,
    link_user_theme_settings_edit
)
from .permissions import (
    permission_theme_delete, permission_theme_edit, permission_theme_view
)


class AppearanceApp(MayanAppConfig):
    app_namespace = 'appearance'
    app_url = 'appearance'
    has_javascript_translations = True
    has_static_media = True
    has_tests = True
    name = 'mayan.apps.appearance'
    static_media_ignore_patterns = (
        'AUTHORS*', 'CHANGE*', 'CONTRIBUT*', 'CODE_OF_CONDUCT*', 'Grunt*',
        'MAINTAIN*', 'README*', '*.less', '*.md', '*.nupkg', '*.nuspec',
        '*.scss*', '*.sh', '*tests*', 'bower*', 'composer.json*',
        'demo*', 'grunt*', 'gulp*', 'install', 'less', 'package.json*',
        'package-lock*', 'test', 'tests', 'variable*',
        'appearance/node_modules/@fancyapps/fancybox/docs/*',
        'appearance/node_modules/@fancyapps/fancybox/src/*',
        'appearance/node_modules/@fortawesome/fontawesome-free/svgs/*',
        'appearance/node_modules/bootswatch/docs/*',
        'appearance/node_modules/jquery/src/*',
        'appearance/node_modules/jquery-form/_config.yml',
        'appearance/node_modules/jquery-form/form.jquery.json',
        'appearance/node_modules/jquery-form/docs/*',
        'appearance/node_modules/jquery-form/src/*',
        'appearance/node_modules/select2/src/*',
        'appearance/node_modules/toastr/karma.conf.js',
        'appearance/node_modules/toastr/toastr.js',
        'appearance/node_modules/toastr/toastr-icon.png',
        'appearance/node_modules/toastr/nuget/*'
    )
    verbose_name = _('Appearance')

    def ready(self):
        super().ready()

        Theme = self.get_model(model_name='Theme')

        User = get_user_model()

        ModelCopy(
            bind_link=True, model=Theme, register_permission=True
        ).add_fields(
            field_names=('label', 'stylesheet',)
        )

        EventModelRegistry.register(model=Theme)

        ModelEventType.register(
            event_types=(
                event_theme_edited,
            ), model=Theme
        )
        ModelEventType.register(
            event_types=(
                event_user_theme_settings_edited,
            ), model=User
        )

        ModelPermission.register(
            model=Theme, permissions=(
                permission_acl_edit, permission_acl_view,
                permission_theme_delete, permission_theme_edit,
                permission_theme_view
            )
        )

        SourceColumn(
            attribute='label', is_identifier=True, is_sortable=True,
            source=Theme
        )

        post_save.connect(
            dispatch_uid='appearance_handler_user_theme_setting_create',
            receiver=handler_user_theme_setting_create,
            sender=settings.AUTH_USER_MODEL
        )

        # Theme

        menu_object.bind_links(
            links=(
                link_theme_delete, link_theme_edit
            ), sources=(Theme,)
        )
        menu_return.bind_links(
            links=(link_theme_list,),
            sources=(
                Theme, 'appearance:theme_list', 'appearance:theme_create'
            )
        )
        menu_secondary.bind_links(
            links=(link_theme_create,),
            sources=(
                Theme, 'appearance:theme_list', 'appearance:theme_create'
            )
        )
        menu_setup.bind_links(
            links=(link_theme_setup,)
        )

        # User

        menu_list_facet.bind_links(
            links=(link_user_theme_settings_detail,), sources=(User,)
        )
        menu_object.bind_links(
            links=(link_user_theme_settings_edit,), sources=(User,)
        )



from django.db import models

from ...search_models import SearchModel


class SearchInterpreterTestMixin:
    def _create_test_models(self):
        self._TestModel = self._create_test_model(
            fields={
                'field_1': models.CharField(
                    max_length=32
                ),
                'field_2': models.CharField(
                    max_length=32
                )
            }, model_name='TestModel'
        )

    def _setup_test_model_search(self):
        self._test_search_model = SearchModel(
            app_label=self._TestModel._meta.app_label,
            model_name=self._TestModel._meta.model_name
        )
        self._test_search_field = self._test_search_model.add_model_field(
            field='field_1'
        )
        self._test_search_field = self._test_search_model.add_model_field(
            field='field_2'
        )



from mayan.apps.acls.models import AccessControlList
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_type_view
)
from mayan.apps.rest_api import generics
from mayan.apps.rest_api.api_view_mixins import ExternalObjectAPIViewMixin

from .models import MetadataType
from .permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove, permission_document_metadata_view,
    permission_metadata_type_create, permission_metadata_type_delete,
    permission_metadata_type_edit, permission_metadata_type_view
)
from .serializers import (
    DocumentMetadataSerializer, DocumentTypeMetadataTypeSerializer,
    MetadataTypeSerializer
)


class APIDocumentMetadataListView(
    ExternalObjectAPIViewMixin, generics.ListCreateAPIView
):
    """
    get: Returns a list of selected document's metadata types and values.
    post: Add an existing metadata type and value to the selected document.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    mayan_external_object_permissions = {
        'GET': (permission_document_metadata_view,),
        'POST': (permission_document_metadata_add,)
    }
    mayan_object_permissions = {
        'GET': (permission_document_metadata_view,)
    }
    ordering_fields = ('id', 'metadata_type', 'value')
    serializer_class = DocumentMetadataSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user,
            'document': self.get_external_object()
        }

    def get_source_queryset(self):
        return self.get_external_object().metadata.all()

    def perform_create(self, serializer):
        if 'metadata_type_id' in serializer.validated_data:
            serializer.validated_data['metadata_type'] = serializer.validated_data['metadata_type_id']

        return super().perform_create(serializer=serializer)


class APIDocumentMetadataView(
    ExternalObjectAPIViewMixin, generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Remove this metadata entry from the selected document.
    get: Return the details of the selected document metadata type and value.
    patch: Edit the selected document metadata type and value.
    put: Edit the selected document metadata type and value.
    """
    external_object_queryset = Document.valid.all()
    external_object_pk_url_kwarg = 'document_id'
    lookup_url_kwarg = 'metadata_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_document_metadata_remove,),
        'GET': (permission_document_metadata_view,),
        'PATCH': (permission_document_metadata_edit,),
        'PUT': (permission_document_metadata_edit,)
    }
    mayan_object_permissions = {
        'DELETE': (permission_document_metadata_remove,),
        'GET': (permission_document_metadata_view,),
        'PATCH': (permission_document_metadata_edit,),
        'PUT': (permission_document_metadata_edit,)
    }
    serializer_class = DocumentMetadataSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }

    def get_source_queryset(self):
        return self.get_external_object().metadata.all()


class APIMetadataTypeListView(generics.ListCreateAPIView):
    """
    get: Returns a list of all the metadata types.
    post: Create a new metadata type.
    """
    mayan_object_permissions = {
        'GET': (permission_metadata_type_view,)
    }
    mayan_view_permissions = {
        'POST': (permission_metadata_type_create,)
    }
    ordering_fields = ('id', 'label', 'name')
    serializer_class = MetadataTypeSerializer
    source_queryset = MetadataType.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIMetadataTypeView(generics.RetrieveUpdateDestroyAPIView):
    """
    delete: Delete the selected metadata type.
    get: Return the details of the selected metadata type.
    patch: Edit the selected metadata type.
    put: Edit the selected metadata type.
    """
    lookup_url_kwarg = 'metadata_type_id'
    mayan_object_permissions = {
        'GET': (permission_metadata_type_view,),
        'PUT': (permission_metadata_type_edit,),
        'PATCH': (permission_metadata_type_edit,),
        'DELETE': (permission_metadata_type_delete,)
    }
    serializer_class = MetadataTypeSerializer
    source_queryset = MetadataType.objects.all()

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class MetadataTypeFilterAPIMixin:
    metadata_type_permissions = {
        'DELETE': (permission_metadata_type_edit,),
        'GET': (permission_metadata_type_view,),
        'PATCH': (permission_metadata_type_edit,),
        'PUT': (permission_metadata_type_edit,)
    }

    def get_metadata_type_queryset(self):
        permission = self.metadata_type_permissions.get(
            self.request.method, (None,)
        )[0]

        queryset = MetadataType.objects.all()
        if permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=permission, queryset=queryset,
                user=self.request.user
            )

        return queryset

    def get_source_queryset(self):
        return self.get_external_object().metadata.filter(
            metadata_type__in=self.get_metadata_type_queryset()
        )


class APIDocumentTypeMetadataTypeListView(
    ExternalObjectAPIViewMixin, MetadataTypeFilterAPIMixin,
    generics.ListCreateAPIView
):
    """
    get: Returns a list of selected document type's metadata types.
    post: Add a metadata type to the selected document type.
    """
    external_object_class = DocumentType
    external_object_pk_url_kwarg = 'document_type_id'
    mayan_external_object_permissions = {
        'GET': (permission_document_type_view,),
        'POST': (permission_document_type_edit,)
    }
    serializer_class = DocumentTypeMetadataTypeSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }


class APIDocumentTypeMetadataTypeView(
    ExternalObjectAPIViewMixin, MetadataTypeFilterAPIMixin,
    generics.RetrieveUpdateDestroyAPIView
):
    """
    delete: Remove a metadata type from a document type.
    get: Retrieve the details of a document type metadata type.
    patch: Edit the selected document type metadata type.
    put: Edit the selected document type metadata type.
    """
    external_object_class = DocumentType
    external_object_pk_url_kwarg = 'document_type_id'
    lookup_url_kwarg = 'metadata_type_id'
    mayan_external_object_permissions = {
        'DELETE': (permission_document_type_edit,),
        'GET': (permission_document_type_view,),
        'PATCH': (permission_document_type_edit,),
        'PUT': (permission_document_type_edit,)
    }
    serializer_class = DocumentTypeMetadataTypeSerializer

    def get_instance_extra_data(self):
        return {
            '_event_actor': self.request.user
        }



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('linking', '0001_initial')
    ]

    operations = [
        migrations.CreateModel(
            name='ResolvedSmartLink',
            fields=[],
            options={
                'proxy': True
            },
            bases=('linking.smartlink',),
        )
    ]



from mayan.apps.common.tests.literals import (
    TEST_ARCHIVE_MSG_STRANGE_DATE_FILENAME, TEST_ARCHIVE_MSG_STRANGE_DATE_PATH
)
from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.source_compressed.source_backends.literals import (
    SOURCE_UNCOMPRESS_CHOICE_ALWAYS, SOURCE_UNCOMPRESS_CHOICE_ASK,
    SOURCE_UNCOMPRESS_CHOICE_NEVER
)

from .mixins import CompressedSourceTestMixin


class CompressedSourceBackendActionDocumentUploadTestCase(
    CompressedSourceTestMixin, GenericDocumentTestCase
):
    _test_source_create_auto = False
    _test_source_file_path = TEST_ARCHIVE_MSG_STRANGE_DATE_PATH
    auto_upload_test_document = False

    def test_compressed_always(self):
        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ALWAYS}
        )

        document_count = Document.objects.count()

        self._test_object_track()

        self._clear_events()

        with open(file=self._test_source_file_path, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={'file_object': file_object}
            )

        self._test_object_set()

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(self._test_document.label, 'message.txt')
        self.assertEqual(self._test_document.file_latest.size, 2711)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_document.file_latest)
        self.assertEqual(events[1].target, self._test_document.file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_document.file_latest)
        self.assertEqual(events[2].target, self._test_document.file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_document.version_active)
        self.assertEqual(events[3].target, self._test_document.version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document.version_active
        )
        self.assertEqual(
            events[4].actor, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    def test_compressed_ask_false(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK}
        )

        document_count = Document.objects.count()

        self._test_object_track()

        self._clear_events()

        with open(file=self._test_source_file_path, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={
                    'expand': False,
                    'file_object': file_object
                }
            )

        self._test_object_set()

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(
            self._test_document.label, TEST_ARCHIVE_MSG_STRANGE_DATE_FILENAME
        )
        self.assertEqual(self._test_document.file_latest.size, 31744)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_document.file_latest)
        self.assertEqual(events[1].target, self._test_document.file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_document.file_latest)
        self.assertEqual(events[2].target, self._test_document.file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_document.version_active)
        self.assertEqual(events[3].target, self._test_document.version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document.version_active
        )
        self.assertEqual(
            events[4].actor, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    def test_compressed_ask_true(self):
        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK}
        )

        document_count = Document.objects.count()

        self._test_object_track()

        self._clear_events()

        with open(file=self._test_source_file_path, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={
                    'expand': True,
                    'file_object': file_object
                }
            )

        self._test_object_set()

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(self._test_document.label, 'message.txt')
        self.assertEqual(self._test_document.file_latest.size, 2711)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_document.file_latest)
        self.assertEqual(events[1].target, self._test_document.file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_document.file_latest)
        self.assertEqual(events[2].target, self._test_document.file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_document.version_active)
        self.assertEqual(events[3].target, self._test_document.version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document.version_active
        )
        self.assertEqual(
            events[4].actor, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    def test_compressed_never(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source_create(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        document_count = Document.objects.count()

        self._test_object_track()

        self._clear_events()

        with open(file=self._test_source_file_path, mode='rb') as file_object:
            self._execute_test_source_action(
                action_name='document_upload',
                extra_data={'file_object': file_object}
            )

        self._test_object_set()

        self.assertEqual(
            Document.objects.count(), document_count + 1
        )

        self.assertEqual(
            self._test_document.label, TEST_ARCHIVE_MSG_STRANGE_DATE_FILENAME
        )
        self.assertEqual(self._test_document.file_latest.size, 31744)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_document)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, self._test_document)
        self.assertEqual(events[1].actor, self._test_document.file_latest)
        self.assertEqual(events[1].target, self._test_document.file_latest)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, self._test_document)
        self.assertEqual(events[2].actor, self._test_document.file_latest)
        self.assertEqual(events[2].target, self._test_document.file_latest)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, self._test_document)
        self.assertEqual(events[3].actor, self._test_document.version_active)
        self.assertEqual(events[3].target, self._test_document.version_active)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(
            events[4].action_object, self._test_document.version_active
        )
        self.assertEqual(
            events[4].actor, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].target, self._test_document.version_active.pages.first()
        )
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.search import search_model_document
from mayan.apps.dynamic_search.search_models import SearchModel

from .permissions import permission_metadata_type_view

# Document

search_model_document.add_model_field(
    field='metadata__metadata_type__name', label=_('Metadata type')
)
search_model_document.add_model_field(
    field='metadata__value', label=_('Metadata value')
)

# Metadata type

search_model_metadata_type = SearchModel(
    app_label='metadata', model_name='MetadataType',
    permission=permission_metadata_type_view,
    serializer_path='mayan.apps.metadata.serializers.MetadataTypeSerializer'
)

search_model_metadata_type.add_model_field(field='default')
search_model_metadata_type.add_model_field(field='label')
search_model_metadata_type.add_model_field(field='lookup')
search_model_metadata_type.add_model_field(field='name')
search_model_metadata_type.add_model_field(field='parser')
search_model_metadata_type.add_model_field(field='validation')



from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_version_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.tests.mixins.document_type_mixins import DocumentTypeTestMixin
from mayan.apps.sources.tasks import task_source_backend_action_execute
from mayan.apps.testing.tests.base import BaseTestCase

from .mixins import PeriodicSourceBackendTestMixin


class SourceTaskTestCase(
    PeriodicSourceBackendTestMixin, DocumentTypeTestMixin, BaseTestCase
):
    def test_task_execute(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        test_document_count = Document.objects.count()

        self._clear_events()

        task_source_backend_action_execute.apply_async(
            kwargs={
                'action_name': 'document_upload',
                'source_id': self._test_source.pk
            }
        )

        self.assertEqual(
            Document.objects.count(), test_document_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 3)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, test_document)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, test_document_file)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, test_document_version)
        self.assertEqual(events[2].target, test_document_version)
        self.assertEqual(events[2].verb, event_document_version_created.id)

    def test_task_execute_disabled(self):
        self._silence_logger(name='mayan.apps.converter.backends')

        self._test_source.enabled = False
        self._test_source.save()

        test_document_count = Document.objects.count()

        self._clear_events()

        task_source_backend_action_execute.apply_async(
            kwargs={
                'action_name': 'document_upload',
                'source_id': self._test_source.pk
            }
        )

        self.assertEqual(
            Document.objects.count(), test_document_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



import os

from django.utils.module_loading import import_string

from mayan.apps.common.tests.mixins import ManagementCommandTestMixin
from mayan.apps.smart_settings.classes import SettingNamespace

from ..exceptions import LockError
from ..literals import COMMAND_NAME_LOCK_MANAGER_PURGE_LOCKS
from ..settings import setting_default_lock_timeout

from .literals import TEST_LOCK_1


class LockBackendManagementCommandTestCaseMixin(ManagementCommandTestMixin):
    _test_management_command_name = COMMAND_NAME_LOCK_MANAGER_PURGE_LOCKS

    def test_purge_locks_command(self):
        self._test_locking_backend.acquire_lock(
            name=TEST_LOCK_1, timeout=20
        )

        # lock_1 not release and not expired, should raise LockError.
        with self.assertRaises(expected_exception=LockError):
            self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        os.environ['MAYAN_LOCK_MANAGER_BACKEND'] = self._test_locking_backend_string
        SettingNamespace.invalidate_cache_all()
        self._call_test_management_command()

        # lock_1 not release but has expired, should not raise LockError.
        lock_2 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_2.release()


class LockBackendTestMixin:
    def setUp(self):
        super().setUp()
        self._test_locking_backend = import_string(
            dotted_path=self._test_locking_backend_string
        )


class LockBackendTestCaseMixin:
    def test_exclusive(self):
        lock_1 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)
        with self.assertRaises(expected_exception=LockError):
            self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_1.release()

    def test_release(self):
        lock_1 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)
        lock_1.release()
        lock_2 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_2.release()

    def test_timeout_expired(self):
        self._test_locking_backend.acquire_lock(name=TEST_LOCK_1, timeout=1)

        # lock_1 not release and not expired, should raise LockError.
        with self.assertRaises(expected_exception=LockError):
            self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        self._test_delay(seconds=1.01)

        # lock_1 not release but has expired, should not raise LockError.
        lock_2 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_2.release()

    def test_double_release(self):
        lock_1 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)
        lock_1.release()

    def test_release_expired(self):
        lock_1 = self._test_locking_backend.acquire_lock(
            name=TEST_LOCK_1, timeout=1
        )

        self._test_delay(seconds=1.01)

        lock_1.release()
        # No exception is raised even though the lock has expired.
        # The logic is that checking for expired locks during release is
        # not necessary as any attempt by someone else to acquire the lock
        # would be successful, even after an extended lapse of time.

    def test_release_expired_reaquired(self):
        self._test_locking_backend.acquire_lock(name=TEST_LOCK_1, timeout=1)

        self._test_delay(seconds=1.01)

        # TEST_LOCK_1 is expired so trying to acquire it should not return
        # an error.
        lock_2 = self._test_locking_backend.acquire_lock(
            name=TEST_LOCK_1, timeout=1
        )

        # Cleanup.
        lock_2.release()

    def test_purge(self):
        self._test_locking_backend.acquire_lock(
            name=TEST_LOCK_1, timeout=30
        )

        # lock_1 not release and not expired, should raise LockError.
        with self.assertRaises(expected_exception=LockError):
            self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        self._test_locking_backend.purge_locks()

        # lock_1 not release but has expired, should not raise LockError.
        lock_2 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_2.release()


class DefaultTimeoutTestMixin:
    def setUp(self):
        super().setUp()
        self._test_locking_backend = import_string(
            dotted_path=self._test_locking_backend_string
        )
        setting_default_lock_timeout.set(value=1)
        self.assertEqual(setting_default_lock_timeout.value, 1)

    def test_default_timeout_expired(self):
        self._test_locking_backend.acquire_lock(
            name=TEST_LOCK_1, timeout=None
        )

        # lock_1 not release and not expired, should raise LockError.
        with self.assertRaises(expected_exception=LockError):
            self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        self._test_delay(seconds=1.01)

        # lock_1 not release but has expired, should not raise LockError.
        lock_2 = self._test_locking_backend.acquire_lock(name=TEST_LOCK_1)

        # Cleanup.
        lock_2.release()



from django.conf.urls import url

from .api_views import APISearchModelList, APISearchView
from .views import (
    SearchAdvancedView, SearchAgainView, SearchBackendReindexView,
    SearchResultsView, SearchSimpleView
)

urlpatterns_search = [
    url(
        regex=r'^again/(?P<search_model_pk>[\.\w]+)/$', name='search_again',
        view=SearchAgainView.as_view()
    ),
    url(
        regex=r'^advanced/(?P<search_model_pk>[\.\w]+)/$',
        name='search_advanced', view=SearchAdvancedView.as_view()
    ),
    url(
        regex=r'^advanced/$', name='search_advanced',
        view=SearchAdvancedView.as_view()
    ),
    url(
        regex=r'^results/$', name='search_results',
        view=SearchResultsView.as_view()
    ),
    url(
        regex=r'^results/(?P<search_model_pk>[\.\w]+)/$',
        name='search_results', view=SearchResultsView.as_view()
    ),
    url(
        regex=r'^simple/(?P<search_model_pk>[\.\w]+)/$',
        name='search_simple', view=SearchSimpleView.as_view()
    )
]

urlpatterns_tools = [
    url(
        regex=r'^backend/reindex/$', name='search_backend_reindex',
        view=SearchBackendReindexView.as_view()
    )
]

urlpatterns = []
urlpatterns.extend(urlpatterns_search)
urlpatterns.extend(urlpatterns_tools)

api_urls = [
    url(
        regex=r'^search/(?P<search_model_pk>[\.\w]+)/$', name='search-view',
        view=APISearchView.as_view()
    ),
    url(
        regex=r'^search/advanced/(?P<search_model_pk>[\.\w]+)/$',
        name='advanced-search-view', view=APISearchView.as_view()
    ),
    url(
        regex=r'^search_models/$', name='searchmodel-list',
        view=APISearchModelList.as_view()
    )
]



from .exceptions import (
    DynamicSearchInterpreterError, DynamicSearchInterpreterUnknownSearchType,
    DynamicSearchScopedQueryError
)
from .literals import MATCH_ALL_FIELD_NAME, SCOPE_MARKER
from .scoped_queries import (
    ScopedQuery, ScopedQueryEntryControlResult, ScopedQueryEntryDataFilter,
    ScopedQueryEntryDataOperator
)
from .search_query_terms import QueryTerm
from .settings import setting_default_operator
from .utils import get_match_all_value


class SearchInterpreter:
    _registry = {}

    @classmethod
    def all(cls):
        sorted_keys = sorted(
            cls._registry.keys()
        )
        return [
            cls._registry[key] for key in sorted_keys
        ]

    @classmethod
    def init(cls, query, search_model, prefix=''):
        """
        Initialization router. Calling this method will cycle all possible
        subclasses and return an instance of the subclass that can handle
        the query type based on the arguments.
        """
        for klass in cls.all():
            checked_query = klass.check(prefix=prefix, query=query)

            if checked_query is not None:
                return klass(
                    query=checked_query, search_model=search_model,
                    prefix=''
                )

        raise DynamicSearchInterpreterUnknownSearchType(
            'No `SearchInterpreter` subclass available that can handle '
            'this search syntax.'
        )

    @classmethod
    def register(cls, klass, priority):
        if priority in cls._registry:
            raise DynamicSearchInterpreterError(
                'Search interpreter `{}` is already registered for '
                'priority `{}`.'.format(
                    cls._registry[priority], priority
                )
            )
        else:
            cls._registry[priority] = klass

    def __init__(self, query, search_model, prefix=''):
        self.prefix = prefix
        self.query = query
        self.scoped_query_class = ScopedQuery
        self.search_model = search_model

        self.search_field_names = [
            search_field.field_name for search_field in self.search_model.search_fields
        ]

    @classmethod
    def do_prefix_remove(cls, prefix, value):
        if value.startswith(prefix):
            return value[len(prefix):]

    def do_query_decode(self, query=None):
        return self._do_query_decode(query=query)

    def do_resolve(self, search_backend):
        scoped_query = self.do_query_decode()

        return scoped_query.do_resolve(search_backend=search_backend)

    def get_scoped_query_instance(self):
        return self.scoped_query_class(search_model=self.search_model)

    @property
    def is_empty(self):
        return self.do_query_decode().is_empty

    def to_explain(self):
        """
        Generate a human readable version of the query.
        """
        clean_query = self.do_query_cleanup()

        scoped_query = self.do_query_decode(query=clean_query)
        return scoped_query.to_explain()

    def to_string(self):
        raise NotImplementedError


class SearchInterpreterAdvanced(SearchInterpreter):
    """
    Search interpreters that decodes user queries using field names and
    values.
    """

    @classmethod
    def check(cls, query, prefix=''):
        result = {}
        for key, value in query.items():
            key = cls.do_prefix_remove(prefix=prefix, value=key)
            if key:
                if not key.startswith(SCOPE_MARKER):
                    result[key] = value

        if result:
            return result

    def _do_query_decode(self, query=None):
        query = query or self.query.copy()

        self.global_and_search = get_match_all_value(
            value=query.pop(MATCH_ALL_FIELD_NAME, 'no')
        )

        if self.global_and_search:
            inter_field_operator = 'AND'
        else:
            inter_field_operator = 'OR'

        query_field_term_dictionary = QueryTerm.do_query_parse(
            query=query
        )

        scoped_query = self.get_scoped_query_instance()

        self.do_scope_search_compose(
            inter_field_operator=inter_field_operator,
            query_field_term_dictionary=query_field_term_dictionary,
            scoped_query=scoped_query
        )

        if self.result_scope is not None:
            scoped_query_entry = ScopedQueryEntryControlResult(
                scoped_query=scoped_query, result_scope_identifier=str(
                    self.result_scope
                )
            )
            scoped_query.do_scope_entry_add(
                scope_entry=scoped_query_entry
            )

        return scoped_query

    def do_resolve(self, search_backend):
        try:
            return super().do_resolve(search_backend=search_backend)
        except DynamicSearchScopedQueryError:
            if not self.result_scope:
                return self.search_model.get_queryset().none()
            else:
                raise

    def do_scope_search_compose(
        self, inter_field_operator, query_field_term_dictionary, scoped_query
    ):
        self.result_scope = None

        field_result_scope_list = []
        scope_id = 0

        for key, term_list in query_field_term_dictionary.items():
            field_scopes = []
            field_operator = setting_default_operator.value

            for term in term_list:
                if term.is_meta:
                    field_operator = term.text
                else:
                    scoped_query_entry = ScopedQueryEntryDataFilter(
                        field_name=key, is_quoted_value=term.is_quoted,
                        is_raw_value=term.is_raw,
                        scope_identifier=str(scope_id),
                        scoped_query=scoped_query, value=term.text
                    )
                    scoped_query.do_scope_entry_add(
                        scope_entry=scoped_query_entry
                    )

                    field_scopes.append(scope_id)
                    scope_id += 1

            if field_scopes:
                new_scope_id = self.do_scope_operators_add(
                    operator_text=field_operator, result_scope=scope_id,
                    scope_id_list=field_scopes, scoped_query=scoped_query
                )
                if new_scope_id is not None:
                    scope_id = new_scope_id + 1

                field_result_scope_list.append(scope_id - 1)

        if field_result_scope_list:
            self.result_scope = self.do_scope_operators_add(
                operator_text=inter_field_operator, result_scope=scope_id,
                scope_id_list=field_result_scope_list,
                scoped_query=scoped_query
            )

    def do_scope_operators_add(
        self, scope_id_list, scoped_query, result_scope, operator_text='AND'
    ):
        """
        Add scope operators in bulk to the specified list of scopes while
        keeping track of the scopes created in the process.
        """
        if scope_id_list:
            operand_left = scope_id_list[0]

            for scope_id in scope_id_list[1:]:
                operand_right = scope_id

                scoped_query_entry = ScopedQueryEntryDataOperator(
                    operand_list=(
                        str(operand_left), str(operand_right)
                    ),
                    operator_text=operator_text,
                    scope_identifier=str(result_scope),
                    scoped_query=scoped_query
                )
                scoped_query.do_scope_entry_add(
                    scope_entry=scoped_query_entry
                )

                operand_left = result_scope
                result_scope += 1

            return result_scope - 1

    def do_query_cleanup(self):
        result = {}

        scoped_query = self.get_scoped_query_instance()

        for key, value in self.query.items():
            key = self.do_prefix_remove(prefix=self.prefix, value=key)

            if key in scoped_query.search_model.search_field_name_list or key == MATCH_ALL_FIELD_NAME:
                result[key] = value

        return result

    def to_string(self):
        scoped_query = self.do_query_decode()
        return scoped_query.to_string()


class SearchInterpreterScoped(SearchInterpreter):
    @classmethod
    def check(cls, query, prefix=''):
        result = {}
        for key, value in query.items():
            key = cls.do_prefix_remove(prefix=prefix, value=key)
            if key:
                if key.startswith(SCOPE_MARKER):
                    result[key] = value

        if result:
            return result

    def _do_query_decode(self, query=None):
        """
        Converts a user scoped query into an internal scope query
        collection.
        """
        query = query or self.query
        scoped_query = self.get_scoped_query_instance()

        for key, value in query.items():
            scoped_query.do_scope_entry_init(key=key, value=value)

        return scoped_query

    def do_query_cleanup(self):
        result = {}
        scoped_query = self.get_scoped_query_instance()

        for key, value in self.query.items():
            key = self.do_prefix_remove(prefix=self.prefix, value=key)
            if key:
                try:
                    scoped_query.do_scope_entry_init(
                        key=key, value=value
                    )
                except DynamicSearchScopedQueryError:
                    """Ignore and continue"""
                else:
                    result[key] = value

        return result

    def to_string(self):
        scoped_query = self.do_query_decode()
        return scoped_query.to_string()


SearchInterpreter.register(klass=SearchInterpreterAdvanced, priority=1)
SearchInterpreter.register(klass=SearchInterpreterScoped, priority=0)



from rest_framework import status

from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.django_gpg.tests.literals import TEST_KEY_PUBLIC_ID
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_detached_signature_created, event_detached_signature_deleted,
    event_detached_signature_uploaded
)
from ..permissions import (
    permission_document_file_sign_detached,
    permission_document_file_signature_delete,
    permission_document_file_signature_view,
    permission_document_file_signature_upload
)

from .mixins import (
    DetachedSignatureAPIViewTestMixin, DetachedSignatureTestMixin
)


class DetachedSignatureDocumentAPIViewTestCase(
    DocumentTestMixin, DetachedSignatureAPIViewTestMixin,
    KeyTestMixin, DetachedSignatureTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def test_document_signature_detached_delete_api_view_no_permission(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        signatures = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_signature_detached_delete_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_delete_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_delete
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_delete_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_document_file)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_deleted.id)

    def test_trashed_document_signature_detached_delete_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_delete
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_detached_delete_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_detail_api_view_no_permission(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self._clear_events()

        response = self._request_test_document_signature_detached_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_detail_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_detail_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            response.data['key_id'], TEST_KEY_PUBLIC_ID
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_detached_detail_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_detached_detail_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_list_api_view_no_permission(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self._clear_events()

        response = self._request_test_document_signature_detached_list_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_list_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_list_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['key_id'], TEST_KEY_PUBLIC_ID
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_signature_detached_list_api_view_with_access(self):
        self._upload_test_document()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_detached_list_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_sign_api_view_with_no_permission(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_signature_detached_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_sign_api_view_with_document_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_sign_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_sign_api_view_with_key_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_sign_api_view_with_full_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_sign_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_document.file_latest.signatures.first().detachedsignature
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_created.id)

    def test_trashed_document_signature_detached_sign_api_view_with_full_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self._test_key_private,
            permission=permission_key_sign
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_detached_sign_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_upload_api_view_no_permission(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self._clear_events()

        response = self._request_test_document_signature_detached_upload_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_signature_detached_upload_api_view_with_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_upload
        )

        self._clear_events()

        response = self._request_test_document_signature_detached_upload_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(
            events[0].action_object,
            self._test_document.file_latest.signatures.first().detachedsignature
        )
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document_file)
        self.assertEqual(events[0].verb, event_detached_signature_uploaded.id)

    def test_trashed_document_signature_detached_upload_api_view_with_access(self):
        self._upload_test_document()
        self._create_test_key_private()

        signatures = self._test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_file_signature_upload
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_test_document_signature_detached_upload_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.file_latest.signatures.count(),
            signatures
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link

from .icons import (
    icon_check_version, icon_dependency_group_entry_list,
    icon_dependency_group_entry_detail, icon_dependency_licenses,
    icon_dependency_group_list
)
from .permissions import permission_dependencies_view

link_check_version = Link(
    icon=icon_check_version, permissions=(permission_dependencies_view,),
    text=_('Check for updates'), view='dependencies:check_version_view'
)
link_dependency_group_list = Link(
    icon=icon_dependency_group_list,
    permissions=(permission_dependencies_view,), text=_('Groups'),
    view='dependencies:dependency_group_list'
)
link_dependency_group_entry_list = Link(
    args='resolved_object.name', icon=icon_dependency_group_entry_list,
    permissions=(permission_dependencies_view,), text=_('Entries'),
    view='dependencies:dependency_group_entry_list'
)
link_dependency_group_entry_detail = Link(
    args=('resolved_object.dependency_group.name', 'resolved_object.name'),
    icon=icon_dependency_group_entry_detail,
    permissions=(permission_dependencies_view,), text=_('Details'),
    view='dependencies:dependency_group_entry_detail'
)
link_packages_licenses = Link(
    icon=icon_dependency_licenses, text=_('Dependencies licenses'),
    view='dependencies:dependency_licenses_view'
)
link_dependency_tool = Link(
    icon=icon_dependency_group_list,
    permissions=(permission_dependencies_view,), text=_('Dependencies'),
    view='dependencies:dependency_group_list'
)



from mayan.apps.testing.tests.base import BaseTestCase

from ..classes import EventModelRegistry, EventType, ModelEventType
from ..decorators import method_event
from ..event_managers import EventManagerMethodAfter

from .mixins import EventTypeTestMixin


class EventManagerTestCase(EventTypeTestMixin, BaseTestCase):
    def setUp(self):
        super().setUp()
        self._create_test_event_type()
        self._create_test_user()
        self._create_test_object()

        EventModelRegistry.register(model=self.TestModel)

        ModelEventType.register(
            event_types=(self._test_event_type,), model=self.TestModel
        )

        EventType.refresh()

    def test_event_ignore(self):
        def method_1(self):
            self._event_ignore = True
            self.method_2()

        @method_event(
            event_manager_class=EventManagerMethodAfter,
            event=self._test_event_type,
            target='self'
        )
        def method_2(self):
            """Nothing"""

        self.TestModel.method_1 = method_1
        self.TestModel.method_2 = method_2

        self._clear_events()

        self._test_object.method_1()

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.testing.tests.base import MayanMigratorTestCase

from ..source_backends import SourceBackendSANEScanner


class SourceBackendPathMigrationTestCase(MayanMigratorTestCase):
    migrate_from = ('sources', '0028_auto_20210905_0558')
    migrate_to = ('source_sane_scanners', '0003_fix_backend_paths')

    def prepare(self):
        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        Source.objects.create(
            backend_path='mayan.apps.sources.source_backends.sane_scanner_backends.SourceBackendSANEScanner',
            label='test source SANE'
        )

    def test_source_backend_path_updates(self):
        Source = self.old_state.apps.get_model(
            app_label='sources', model_name='Source'
        )

        self.assertEqual(
            Source.objects.get(label='test source SANE').backend_path,
            SourceBackendSANEScanner.get_class_path()
        )



from .search_models import SearchModel


class QuerysetSearchModelMixin:
    def get_search_model_from_queryset(self, queryset):
        try:
            model = queryset.model
        except AttributeError:
            return
        else:
            try:
                return SearchModel.get_for_model(instance=model)
            except KeyError:
                return



import os

from django import forms as django_forms
from django.apps import apps
from django.conf import settings
from django.contrib.admin.utils import (
    get_fields_from_path, help_text_for_field, label_for_field
)
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.forms import Form as DjangoForm, ModelForm as DjangoModelForm
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.utils import resolve_attribute

from .form_mixins import FormMixinDynamicFields, FormMixinFieldsets
from .form_options import DetailFormOption, FilteredSelectionFormOptions
from .widgets import DisableableSelectWidget, PlainWidget, TextAreaDiv


class Form(FormMixinFieldsets, DjangoForm):
    """Mayan's default form class."""


class ModelForm(FormMixinFieldsets, DjangoModelForm):
    """Mayan's default model form class."""


class ChoiceForm(Form):
    """
    Form to be used in side by side templates used to add or remove
    items from a many to many field.
    """
    search = django_forms.CharField(
        label=_('Search'), required=False,
        widget=django_forms.widgets.TextInput(
            attrs={
                'autocomplete': 'off',
                'class': 'views-select-search',
                'placeholder': 'Filter list'
            }
        )
    )
    selection = django_forms.MultipleChoiceField(
        required=False, widget=DisableableSelectWidget()
    )

    def __init__(self, *args, **kwargs):
        choices = kwargs.pop(
            'choices', []
        )
        label = kwargs.pop(
            'label', _('Selection')
        )
        help_text = kwargs.pop('help_text', None)
        disabled_choices = kwargs.pop(
            'disabled_choices', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['selection'].choices = choices
        self.fields['selection'].label = label
        self.fields['selection'].help_text = help_text
        self.fields['selection'].widget.disabled_choices = disabled_choices
        self.fields['selection'].widget.attrs.update(
            {
                'class': 'full-height input-hotkey-double-click',
                'data-height-difference': '495'
            }
        )


class DetailForm(ModelForm):
    def __init__(self, *args, **kwargs):
        self.opts = DetailFormOption(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )
        super().__init__(*args, **kwargs)

        for field_index, extra_field in enumerate(iterable=self.opts.extra_fields):
            obj = extra_field.get('object', self.instance)
            field = extra_field.get('field', None)
            func = extra_field.get('func', None)
            label = extra_field.get('label', None)
            help_text = extra_field.get('help_text', None)

            if field:
                if not label:
                    # If label is not specified try to get it from the object
                    # itself.
                    try:
                        fields = get_fields_from_path(model=obj, path=field)
                    except FieldDoesNotExist:
                        # Might be property of a method.
                        label = getattr(
                            getattr(obj, field), 'short_description',
                            field
                        )
                    else:
                        label = label_for_field(
                            model=obj, name=fields[-1].name
                        )

                if not help_text:
                    # If help_text is not specified try to get it from the
                    # object itself.
                    try:
                        fields = get_fields_from_path(model=obj, path=field)
                    except FieldDoesNotExist:
                        # Might be property of a method.
                        help_text = getattr(
                            getattr(obj, field), 'help_text', None
                        )
                    else:
                        help_text = help_text_for_field(
                            model=obj, name=fields[-1].name
                        )

                value = resolve_attribute(attribute=field, obj=obj)

            if func:
                value = func(obj)

            field = field or 'anonymous_field_{}'.format(field_index)

            if isinstance(value, models.query.QuerySet):
                self.fields[field] = django_forms.ModelMultipleChoiceField(
                    queryset=value, label=label
                )
            else:
                self.fields[field] = django_forms.CharField(
                    initial=value, label=label, help_text=help_text,
                    widget=extra_field.get('widget', PlainWidget)
                )

        for field_name, field in self.fields.items():
            self.fields[field_name].widget.attrs.update(
                {'readonly': 'readonly'}
            )


class DynamicForm(FormMixinDynamicFields, Form):
    """Normal dynamic form."""


class DynamicModelForm(FormMixinDynamicFields, ModelForm):
    """Dynamic model form."""


class FileDisplayForm(Form):
    DIRECTORY = None
    FILENAME = None

    text = django_forms.CharField(
        label='',
        widget=TextAreaDiv(
            attrs={
                'class': 'full-height scrollable',
                'data-height-difference': 270
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.DIRECTORY or self.FILENAME:
            file_path = os.path.join(
                settings.BASE_DIR, os.sep.join(self.DIRECTORY), self.FILENAME
            )
            with open(file=file_path) as file_object:
                self.fields['text'].initial = file_object.read()


class FilteredSelectionForm(Form):
    """
    Form to select the from a list of choice filtered by access. Can be
    configured to allow single or multiple selection.
    """
    def __init__(self, *args, **kwargs):
        opts = FilteredSelectionFormOptions(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )

        if opts.queryset is None:
            if not opts.model:
                raise ImproperlyConfigured(
                    '{} requires a queryset or a model to be specified as '
                    'a meta option or passed during initialization.'.format(
                        self.__class__.__name__
                    )
                )

            queryset = opts.model.objects.all()
        else:
            queryset = opts.queryset

        if opts.allow_multiple:
            extra_kwargs = {}
            field_class = django_forms.ModelMultipleChoiceField
            widget_class = django_forms.widgets.SelectMultiple
        else:
            extra_kwargs = {'empty_label': None}
            field_class = django_forms.ModelChoiceField
            widget_class = django_forms.widgets.Select

        if opts.widget_class:
            widget_class = opts.widget_class

        if opts.permission:
            AccessControlList = apps.get_model(
                app_label='acls', model_name='AccessControlList'
            )
            queryset = AccessControlList.objects.restrict_queryset(
                permission=opts.permission, queryset=queryset,
                user=opts.user
            )

        super().__init__(*args, **kwargs)

        self.fields[opts.field_name] = field_class(
            help_text=opts.help_text, label=opts.label,
            queryset=queryset, required=opts.required,
            widget=widget_class(attrs=opts.widget_attributes),
            **extra_kwargs
        )


class RelationshipForm(Form):
    def __init__(self, *args, **kwargs):
        self._event_actor = kwargs.pop('_event_actor')
        super().__init__(*args, **kwargs)

        self.fields['label'] = django_forms.CharField(
            label=_('Label'), required=False,
            widget=django_forms.TextInput(
                attrs={'readonly': 'readonly'}
            )
        )
        self.fields['relationship_type'] = django_forms.ChoiceField(
            choices=self.RELATIONSHIP_CHOICES, label=_('Relationship'),
            widget=django_forms.RadioSelect()
        )

        self.sub_object = self.initial.get('sub_object')
        if self.sub_object:
            self.fields['label'].initial = str(self.sub_object)

            self.initial_relationship_type = self.get_relationship_type()

            self.fields['relationship_type'].initial = self.initial_relationship_type

    def get_new_relationship_instance(self):
        related_manager = getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        )
        main_field_name = related_manager.field.name

        return related_manager.model(
            **{
                main_field_name: self.initial.get('object'),
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_queryset_relationship(self):
        return getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        ).filter(
            **{
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_relationship_instance(self):
        queryset_relationship = self.get_queryset_relationship()
        if queryset_relationship.exists():
            return queryset_relationship.get()
        else:
            return self.get_new_relationship_instance()

    def save(self):
        if self.sub_object:
            if self.cleaned_data['relationship_type'] != self.initial_relationship_type:
                save_method = getattr(
                    self, 'save_relationship_{}'.format(
                        self.cleaned_data['relationship_type']
                    )
                )
                save_method()



import logging

from mayan.apps.acls.models import AccessControlList
from mayan.apps.navigation.classes import Link

from ..exceptions import SourceActionExceptionUnknown
from ..icons import icon_upload_view_link
from ..links import factory_conditional_active_by_source
from ..menus import menu_sources
from ..models import Source

logger = logging.getLogger(name=__name__)


class SourceActionViewMixin:
    object_permission = None
    view_source_action = None

    def dispatch(self, request, *args, **kwargs):
        self.queryset_source_valid = self.get_queryset_source_valid()

        return super().dispatch(request=request, *args, **kwargs)

    def get_queryset_source_valid(self):
        def generator():
            view_source_action = self.get_view_source_action()

            queryset = Source.objects.filter(enabled=True)

            for source in queryset:
                try:
                    action = source.get_action(name=view_source_action)
                    if action.has_interface(interface_name='View'):
                        yield source.pk
                except SourceActionExceptionUnknown:
                    """
                    Source does not support the action, don't add it to the
                    id list of sources.
                    """

        queryset = Source.objects.filter(
            pk__in=generator()
        )

        if self.object_permission:
            queryset = AccessControlList.objects.restrict_queryset(
                permission=self.object_permission, queryset=queryset,
                user=self.request.user
            )

        return queryset

    def get_source_link_queryset(self):
        return self.source_link_queryset


class SourceLinkNavigationViewMixin:
    source_link_action = None
    source_link_permission = None
    source_link_queryset = None

    @staticmethod
    def get_source_link_for(source, view_name, view_kwargs=None):
        kwargs = {
            'source_id': str(source.pk)
        }

        if view_kwargs:
            kwargs.update(**view_kwargs)

        return Link(
            conditional_active=factory_conditional_active_by_source(
                source=source
            ), icon=icon_upload_view_link, keep_query=True,
            kwargs=kwargs, remove_from_query=['page'],
            text=source.label, view=view_name
        )

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        source_link_list = self.get_source_link_list()
        menu_sources.bound_links[
            'sources:document_upload'
        ] = source_link_list
        menu_sources.bound_links[
            'sources:document_file_upload'
        ] = source_link_list
        menu_sources.bound_links[
            'sources:source_action'
        ] = source_link_list

        return context

    def get_source_link_action(self):
        return self.source_link_action

    def get_source_link_list(self):
        queryset_sources = AccessControlList.objects.restrict_queryset(
            permission=self.get_source_link_permission(),
            queryset=self.get_source_link_queryset(),
            user=self.request.user
        )

        source_link_action = self.get_source_link_action()
        for source in queryset_sources:
            try:
                if source.get_action(name=source_link_action):
                    yield SourceLinkNavigationViewMixin.get_source_link_for(
                        source=source,
                        view_kwargs=self.get_source_link_view_kwargs(),
                        view_name=self.get_source_link_view_name()
                    )
            except SourceActionExceptionUnknown:
                """
                Source does not support the action. Ignore attempting
                to add the links.
                """

    def get_source_link_permission(self):
        return self.source_link_permission

    def get_source_link_queryset(self):
        return self.source_link_queryset

    def get_source_link_view_kwargs(self):
        return {}

    def get_source_link_view_name(self):
        return self.source_link_view_name



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_SIGNATURES_BACKEND, DEFAULT_DEFAULT_GPG_PATH,
    DEFAULT_SIGNATURES_KEYSERVER
)

namespace = SettingNamespace(
    label=_('Signatures'), name='django_gpg'
)

setting_gpg_backend = namespace.add_setting(
    default=DEFAULT_SIGNATURES_BACKEND,
    global_name='SIGNATURES_BACKEND',
    help_text=_(
        'Full path to the backend to be used to handle keys and signatures.'
    )
)
setting_gpg_backend_arguments = namespace.add_setting(
    default=DEFAULT_DEFAULT_GPG_PATH,
    global_name='SIGNATURES_BACKEND_ARGUMENTS',
)
setting_keyserver = namespace.add_setting(
    default=DEFAULT_SIGNATURES_KEYSERVER, global_name='SIGNATURES_KEYSERVER',
    help_text=_('Keyserver used to query for keys.')
)



import functools

from .literals import EVENT_MANAGER_ORDER_AFTER, EVENT_MANAGER_ORDER_BEFORE


def method_event(event_manager_class, **event_manager_kwargs):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            event_manager = event_manager_class(
                instance=self, **event_manager_kwargs
            )
            event_manager.prepare()
            # Gather everything set before the wrapped function.
            event_manager.pop_event_attributes()

            if event_manager.order == EVENT_MANAGER_ORDER_BEFORE:
                event_manager.commit()

            result = func(self, *args, **kwargs)

            # Call `pop_event_attributes` again to gather anything else
            # set inside the wrappedd function itself.
            event_manager.pop_event_attributes()

            if event_manager.order == EVENT_MANAGER_ORDER_AFTER:
                event_manager.commit()

            return result
        return wrapper
    return decorator



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('duplicates', '0009_auto_20210124_0738')
    ]

    operations = [
        migrations.AlterField(
            model_name='duplicatebackendentry',
            name='documents',
            field=models.ManyToManyField(
                related_name='as_duplicate', to='documents.Document',
                verbose_name='Duplicated documents'
            )
        )
    ]



from pathlib import Path
from unittest import skip

from django.test import override_settings

from mayan.apps.events.classes import EventType
from mayan.apps.storage.utils import TemporaryDirectory
from mayan.apps.testing.tests.base import BaseTransactionTestCase
from mayan.settings.literals import (
    DEFAULT_USER_SETTINGS_FOLDER, SECRET_KEY_FILENAME, SYSTEM_DIR
)

from ..literals import (
    COMMAND_NAME_COMMON_INITIAL_SETUP, COMMAND_NAME_COMMON_PERFORM_UPGRADE
)

from .mixins import ManagementCommandTestMixin


@skip(
    reason='Skip until existing database persistence with transaction '
    'handling is achieve.'
)
class CommonAppInitialSetupManagementCommandTestCase(
    ManagementCommandTestMixin, BaseTransactionTestCase
):
    _test_management_command_name = COMMAND_NAME_COMMON_INITIAL_SETUP

    def test_command_initial_setup_no_files(self):
        with TemporaryDirectory() as path_name:
            path_temporary_media = Path(path_name, 'media')
            with override_settings(MEDIA_ROOT=str(path_temporary_media)):
                self._test_management_command_name()

            self.assertTrue(
                (path_temporary_media / SYSTEM_DIR).exists()
            )
            self.assertTrue(
                (
                    path_temporary_media / SYSTEM_DIR / SECRET_KEY_FILENAME
                ).exists()
            )
            self.assertTrue(
                (
                    path_temporary_media / DEFAULT_USER_SETTINGS_FOLDER
                ).exists()
            )
            self.assertTrue(
                (
                    path_temporary_media / DEFAULT_USER_SETTINGS_FOLDER / '__init__.py'
                ).exists()
            )

    def test_command_initial_setup_existing_files(self):
        with TemporaryDirectory() as path_name:
            path_temporary_media = Path(path_name, 'media')
            with override_settings(MEDIA_ROOT=str(path_temporary_media)):
                self._test_management_command_name()
                with self.assertRaises(expected_exception=SystemExit):
                    self._test_management_command_name()


class CommonAppPerformUpgradeManagementCommandTestCase(
    ManagementCommandTestMixin, BaseTransactionTestCase
):
    _test_management_command_name = COMMAND_NAME_COMMON_PERFORM_UPGRADE

    def setUp(self):
        super().setUp()
        EventType.refresh()

    def test_command_perform_upgrade_no_files(self):
        with TemporaryDirectory() as path_name:
            path_temporary_media = Path(path_name, 'media')
            with override_settings(MEDIA_ROOT=str(path_temporary_media)):
                with self.assertRaises(expected_exception=FileNotFoundError):
                    self._call_test_management_command(no_dependencies=True)

    def test_command_perform_upgrade_existing_files(self):
        with TemporaryDirectory() as path_name:
            path_temporary_media = Path(path_name, 'media')
            with override_settings(MEDIA_ROOT=str(path_temporary_media)):
                self._test_management_command_name = COMMAND_NAME_COMMON_INITIAL_SETUP
                self._call_test_management_command(no_dependencies=True)
                self._test_management_command_name = COMMAND_NAME_COMMON_PERFORM_UPGRADE
                self._call_test_management_command(no_dependencies=True)



from django_celery_beat.models import IntervalSchedule, PeriodicTask

from ..classes import CeleryQueue, Worker

from .literals import (
    TEST_INTERVAL_SCHEDULE_EVERY, TEST_INTERVAL_SCHEDULE_PERIOD,
    TEST_PERIODIC_TASK_NAME, TEST_PERIODIC_TASK_TASK, TEST_QUEUE_LABEL,
    TEST_QUEUE_NAME, TEST_WORKER_NAME
)


class TaskManagerManagementCommandTestMixin:
    def setUp(self):
        super().setUp()

        self._test_interval_schedule = IntervalSchedule.objects.create(
            every=TEST_INTERVAL_SCHEDULE_EVERY,
            period=TEST_INTERVAL_SCHEDULE_PERIOD
        )
        PeriodicTask.objects.create(
            interval=self._test_interval_schedule,
            name=TEST_PERIODIC_TASK_NAME, task=TEST_PERIODIC_TASK_TASK
        )


class TaskManagerTestMixin:
    def setUp(self):
        super().setUp()
        self._test_queue_list = []
        self._test_task_type_list = []
        self._test_worker_list = []

    def tearDown(self):
        for test_queue in self._test_queue_list:
            test_queue.remove()

    def _create_test_queue(self, label=None, name=None):
        total_test_queue_list = len(self._test_queue_list)
        label = label or '{}_{}'.format(
            TEST_QUEUE_LABEL, total_test_queue_list
        )
        name = name or '{}_{}'.format(TEST_QUEUE_NAME, total_test_queue_list)

        self._test_queue = CeleryQueue(
            label=label, name=name, worker=self._test_worker
        )
        self._test_queue_list.append(self._test_queue)

    def _create_test_task_type(self):
        self._test_task_type = self._test_queue.add_task_type(
            dotted_path=TEST_PERIODIC_TASK_TASK, label='test task type'
        )

    def _create_test_worker(self):
        self._test_worker = Worker(name=TEST_WORKER_NAME)
        self._test_worker_list.append(self._test_worker)


class TaskManagerViewTestMixin:
    def _request_queue_task_type_list(self):
        return self.get(
            viewname='task_manager:queue_task_type_list', kwargs={
                'queue_name': self._test_queue.name
            }
        )

    def _request_worker_list(self):
        return self.get(
            viewname='task_manager:worker_list'
        )

    def _request_worker_queue_list(self):
        return self.get(
            viewname='task_manager:worker_queue_list', kwargs={
                'worker_name': self._test_worker.name
            }
        )



from mayan.apps.testing.tests.base import BaseTestCase

from .mixins.workflow_template_mixins import WorkflowTemplateTestMixin


class WorkflowTemplateModelTestCase(WorkflowTemplateTestMixin, BaseTestCase):
    def test_workflow_template_preview_api_url(self):
        self._create_test_workflow_template()
        self.assertTrue(
            self._test_workflow_template.get_api_image_url()
        )

    def test_workflow_template_preview_generate_image(self):
        self._create_test_workflow_template()
        self.assertTrue(
            self._test_workflow_template.generate_image()
        )
